import os

import matplotlib.pyplot as plt
import dhnx


os.chdir(os.path.dirname(__file__))


# Initialize thermal network
network = dhnx.network.ThermalNetwork()
network = network.from_csv_folder("twn_data")

# Load investment parameter
invest_opt = dhnx.input_output.load_invest_options("invest_data")

# Execute investment optimization
network.optimize_investment(invest_options=invest_opt, write_lp_file=True)

# ####### Postprocessing and Plotting ###########
# Draw network
static_map = dhnx.plotting.StaticMap(network)
static_map.draw(background_map=False)
plt.title("Given network")
plt.show()

# get results
results_edges = network.results.optimization["components"]["pipes"]
print("*Results*")
print(results_edges)
print("")
print("Objective Value: ", network.results.optimization["oemof_meta"]["objective"])

# manually recalculate total costs
total_costs = (33 * 3.162 + 15 * 1 + 18 * 1 + 18 * 0.5) * 0.5
print("Costs re-calculation: ", total_costs)

# get indices which are existing or invested
ind = results_edges[results_edges["capacity"] > 0].index

# select invested edges
network_result = network
network_result.components["pipes"] = results_edges.loc[ind]

# plot results network
static_map = dhnx.plotting.StaticMap(network_result)
static_map.draw(background_map=False)
plt.title("Optimization result")
plt.show()



# -*- coding: utf-8

"""
This script calculates the pressure losses and the needed power for the pump
in a simple tree network.

The assumption of three fittings has been made:
1. Tee connector at the fork
2. Valve at the consumer 1
3. Valve at the consumer 2
"""

import math
import os

import numpy as np
import pandas as pd

path_file = os.path.dirname(__file__)
path = os.path.abspath(os.path.join(path_file, os.pardir, os.pardir, os.pardir))
input_data = os.path.join(path, "examples", "simulation", "tree")
result_path = os.path.join(path_file, "sequences")

if not os.path.exists(result_path):
    os.mkdir(result_path)


def read_data(input_value):
    r"""
    This function is reading the data of a csv with a name given as input value
    """
    return pd.read_csv(os.path.join(input_data, input_value + ".csv"), index_col=0)


# Read input data for every csv component
consumers = read_data("consumers")
pipes = read_data("pipes")
forks = read_data("forks")
producers = read_data("producers")
mass_flow = pd.read_csv(input_data + "/sequences/consumers-mass_flow.csv")

# Constants for calculation
rho = 971.78  # [kg/m3] TODO: later density could calculated with CoolProp
epsilon = 0.01  # [mm]
zeta_tee_connect = (
    0.75  # [-] Estimate of Tee connector: WJ Beek - Transport Phenomena (1999)
)
zeta_tee_separation = (
    2  # [-] Estimate of Tee connector: WJ Beek - Transport Phenomena (1999)
)
zeta_valve = 3.3  # [-] VDI Wärmeatlas für Nennweite von 50 mm
mu = 0.35  # [mPa*s]
eta_pump = 1  # [-]
g = 9.81  # [m/s2]
pi = math.pi

# Initialize variables of type dataframe (needed for later calculations)
(
    v,
    re,
    lambda_simp,
    lambda_adv,
    dp_diss,
    dp_loc,
    dp_loc_tee_i,
    dp_loc_tee_r,
    dp_loc_valve,
    dp_hyd,
    dp,
) = [pd.DataFrame() for variable in range(11)]

# Adjust mass flows to a dataframe containing all mass flows in correct order
# Get mass flows of all consumers
mass_flow_total = mass_flow.iloc[:, 1:]
# Rename the columns to pipes naming convention
mass_flow_total.columns = ["1", "2"]
# Calculate producer mass flow as sum of consumer mass flows
mass_flow_total["0"] = mass_flow_total["1"] + mass_flow_total["2"]
# Change order of columns for later calculation
mass_flow_total = mass_flow_total[["0", "1", "2"]]


for index, node in enumerate(mass_flow_total):
    # Calculation of the velocity v
    v[str(index)] = (
        4
        * mass_flow_total[str(node)]
        / (rho * pi * (pipes["diameter"].iloc[index] / 1000) ** 2)
    )

    # Calculation of Re number
    re[str(index)] = (
        pipes["diameter"].iloc[index] / 1000 * v[str(index)] * rho / (mu / 1000)
    )
    # Calculation of lambda with simple approach
    lambda_simp[str(index)] = (
        0.07
        * re[str(index)] ** -0.13
        * (pipes["diameter"].iloc[index] / 1000) ** -0.14
    )
    # Calculation of lambda with advanced approach
    lambda_adv[str(index)] = (
        1.325
        / np.log(
            epsilon / (1000 * 3.7 * (pipes["diameter"].iloc[index] / 1000))
            + 5.74 / (re[str(index)] ** 0.9)
        )
        ** 2
    )
    # Calculate distributed pressure losses with Darcy-Weissbach-equation
    dp_diss[str(index)] = (
        lambda_simp[str(index)]
        * rho
        * pipes["length"].iloc[index]
        * v[str(index)] ** 2
        / (2 * (pipes["diameter"].iloc[index] / 1000))
    )
    # Calculate local pressure losses resulted from separating Tee (T-Stück) -> i - inlet
    # Localized Pressure losses only occur in the outlet pipes of the tee separator
    if node == "0":
        dp_loc_tee_i[str(index)] = (
            0 * zeta_tee_separation * v[str(index)] ** 2 * rho / 2
        )
    elif node != "0":
        dp_loc_tee_i[str(index)] = zeta_tee_separation * v[str(index)] ** 2 * rho / 2
    # Calculate local pressure losses resulted from connecting Tee (T-Stück) -> r - return
    # Localized Pressure losses only occur in the outlet pipes of the tee connector
    if node == "0":
        dp_loc_tee_r[str(index)] = zeta_tee_connect * v[str(index)] ** 2 * rho / 2
    elif node != "0":
        dp_loc_tee_r[str(index)] = 0 * zeta_tee_connect * v[str(index)] ** 2 * rho / 2
    # Calculate local pressure losses resulted from consumer valves
    if node == "0":
        dp_loc_valve[str(index)] = (
            0 * v[str(index)] ** 2 * rho / 2
        )  # Producer no valve, zeta = 0
    elif node != "0":
        dp_loc_valve[str(index)] = zeta_valve * v[str(index)] ** 2 * rho / 2

# Calculate distributed pressure losses for inlet and return direction (dp_diss * 2)
dp_diss = dp_diss * 2

# Calculate sum of local pressure losses
dp_loc_tee = dp_loc_tee_i + dp_loc_tee_r
dp_loc = dp_loc_tee

# Calculate hydrostatic pressure difference
dp_hyd["0"] = (
    -rho * g * abs(producers["m_over_NHN"][0] - forks["m_over_NHN"][0]) * v["0"] ** 0
)
dp_hyd["1"] = (
    -rho
    * g
    * abs(forks["m_over_NHN"][0] - consumers["m_over_NHN"].iloc[0])
    * v["1"] ** 0
)
dp_hyd["2"] = (
    -rho
    * g
    * abs(forks["m_over_NHN"][0] - consumers["m_over_NHN"].iloc[1])
    * v["2"] ** 0
)

# Calculate total pressure loss
dp = dp_diss + dp_loc + dp_hyd

# Find the consumer with maximum pressure losses
max_pressure_losses = max(dp.loc[:, "1":])

# Calculate global pressure loss
dp_glob = pd.DataFrame(data={"0": np.zeros(len(mass_flow_total))})
dp_glob["0"] = dp_glob["0"] + dp["0"] + dp[max_pressure_losses]

# Calculate pump power
p_el_pump = pd.DataFrame(data={"0": np.zeros(len(mass_flow_total))})

for index, node in enumerate(v):
    p_el_pump["0"] = (
        p_el_pump["0"]
        + 1 / eta_pump * dp[str(index)] / rho * mass_flow_total[str(index)]
    )


# Print results
def parameters():
    r"""
    Writes results in Dictionary

    Returns
    -------
    parameter : dict
                Dictionary with results
    """
    param_dict = {
        "Velocity v [m/s]": v,
        "Reynolds Re [-]": re,
        "Lambda λ (simple approach) [-]": lambda_simp,
        "Lambda λ (advanced approach) [-]": lambda_adv,
        "Distributed pressure losses ∆p_dis [kg/ms2]": dp_diss,
        "Local losses due to tee ∆p_loc,tee [kg/ms2]": dp_loc_tee,
        "Local losses due to valve ∆p_loc,valve [kg/ms2]": dp_loc_valve,
        "Total local losses ∆p_loc [kg/ms2]": dp_loc,
        "Hydrostatic losses ∆p_hyd [kg/ms2]": dp_hyd,
        "Total pressure losses ∆p [kg/ms2]": dp,
        "Global pressure losses ∆p_glob [kg/ms2]": dp_glob,
        "Total electrical energy pump P_el [W]": p_el_pump,
    }
    return param_dict


parameter = parameters()


def print_parameters():
    r"""
    Prints the parameters
    """
    dash = "-" * 60
    print("\n" + dash)
    print("Results at producer (0), consumer 1 (1) and consumer 2 (2)")
    print(dash)

    for name, param in parameter.items():
        print(name + "\n" + str(param) + "\n\n")

    print(dash)


print_parameters()


# Save results to csv
# for name, param in parameter.items():
#    param.insert(0, 'snapshot', np.arange(len(mass_flow_total)))

result_name = [
    "pipes-dist_pressure_losses.csv",
    "pipes_loc_pressure_losses.csv",
    "global-pressure_losses.csv",
    "producers-pump_power.csv",
]
result_list = (
    [list(parameter.keys())[4]]
    + [list(parameter.keys())[5]]
    + list(parameter.keys())[-2:]
)

for index, value in enumerate(result_list):
    parameter[value].to_csv(os.path.join(result_path, result_name[index]), index=False)



__version__ = "0.0.4"

from . import dhn_from_osm  # noqa: F401
from . import graph  # noqa: F401
from . import helpers  # noqa: F401
from . import input_output  # noqa: F401
from . import model  # noqa: F401
from . import network  # noqa: F401
from . import plotting  # noqa: F401
from . import simulation  # noqa: F401
from .gistools import connect_points  # noqa: F401
from .gistools import geometry_operations  # noqa: F401
from .optimization import add_components  # noqa: F401
from .optimization import dhs_nodes  # noqa: F401
from .optimization import oemof_heatpipe  # noqa: F401
from .optimization import optimization_models  # noqa: F401
from .optimization import precalc_hydraulic  # noqa: F401



import os

import dhnx

from . import helpers

# define paths
tests_path = os.path.abspath(os.path.dirname(__file__))

tmp_dir = helpers.extend_basic_path('tmp')


def test_tree_simulation():

    # TODO: Test not working! Nothing happens, if check fails!

    input_data = os.path.join(
        tests_path,
        '_files',
        'tree_network_import')

    expected_results = os.path.join(
        tests_path,
        '_files',
        'tree_network_sim_expected_results',
        'sequences'
    )

    results_path = os.path.join(
        tmp_dir,
        'tree_results'
    )

    # Initialize thermal network
    tnw = dhnx.network.ThermalNetwork()

    # Load data from csv
    tnw.from_csv_folder(input_data)

    # Create simulation model
    tnw.simulate(results_dir=results_path)

    # compare with expected results
    # Workaround since the benchmark data was created in a different way:
    # Ignore differences in columns and index.
    helpers.check_if_csv_dirs_equal(results_path, expected_results, check_arrays=True)


def test_tree_simulation_reverse_pipe_dir():

    # TODO: Test not working! Nothing happens, if check fails!

    input_data = os.path.join(
        tests_path,
        '_files',
        'tree_network_import')

    expected_results = os.path.join(
        tests_path,
        '_files',
        'tree_network_sim_expected_results',
        'sequences'
    )

    results_path = os.path.join(
        tmp_dir,
        'tree_results_reversed_pipe'
    )

    # Initialize thermal network
    tnw = dhnx.network.ThermalNetwork()

    # Load data from csv
    tnw.from_csv_folder(input_data)

    # Swap direction of first pipe
    tnw.components.pipes.loc[0, ['from_node', 'to_node']] = \
        tnw.components.pipes.loc[0, ['to_node', 'from_node']].values

    # Create simulation model
    tnw.simulate(results_dir=results_path)

    # compare with expected results
    # Workaround since the benchmark data was created in a different way:
    # Ignore differences in columns and index.
    helpers.check_if_csv_dirs_equal(results_path, expected_results, check_arrays=True)



import matplotlib.pyplot as plt
import dhnx
import os

os.chdir(os.path.dirname(__file__))
# Initialize thermal network
network = dhnx.network.ThermalNetwork()
network = network.from_csv_folder("twn_data")

# general optimisation settings
set = {
    "num_ts": 3,
    "time_res": 1,
    "start_date": "1/1/2018",
    "frequence": "H",
    "solver": "cbc",
    "solve_kw": {"tee": True},
    "simultaneity": 0.8,
    "bidirectional_pipes": True,
    "print_logging_info": True,
    "heat_demand": "series",
}

invest_opt = dhnx.input_output.load_invest_options("invest_data")

network.optimize_investment(**set, invest_options=invest_opt)

# Draw network
static_map = dhnx.plotting.StaticMap(network)
static_map.draw(background_map=False)
plt.title("Given network")
plt.show()

# get results
results_edges = network.results.optimization["components"]["pipes"]
print("*Results*")
print(results_edges)

# get indices which are existing or invested
ind = results_edges[results_edges["capacity"] > 0].index

edges = network.components["pipes"]
ind_exist = list(edges[edges["existing"] == 1].index)

# plot existing network
network_exist = network
network_exist.components["pipes"] = edges.loc[ind_exist]

# plot existing network
static_map = dhnx.plotting.StaticMap(network_exist)
static_map.draw(background_map=False)
plt.title("Existing pipes")
plt.show()

# select xisting or invested edges
network_result = network
network_result.components["pipes"] = results_edges.loc[ind]

# plot results network
static_map = dhnx.plotting.StaticMap(network_result)
static_map.draw(background_map=False)
plt.title("Investment result")
plt.show()



# -*- coding: utf-8 -*-

"""
Create a district heating network from OpenStreetMap data,
and perform a DHS Investment Optimisation.

Overview
--------

Part I: Get OSM data
Part II: Process the geometry for DHNx
Part III: Initialise the ThermalNetwork and perform the Optimisation
Part IV: Check the results

Contributors:
- Joris Zimmermann
- Johannes Röder
"""
import numpy as np
import osmnx as ox
from shapely import geometry
import matplotlib.pyplot as plt
import os

import logging
from oemof.tools import logger

from dhnx.network import ThermalNetwork
from dhnx.input_output import load_invest_options
from dhnx.gistools.connect_points import process_geometry

logger.define_logging(
    screen_level=logging.INFO,
    logfile="dhnx.log"
)

os.chdir(os.path.dirname(__file__))

# Part I: Get OSM data #############

# select the street types you want to consider as DHS routes
# see: https://wiki.openstreetmap.org/wiki/Key:highway
streets = dict({
    'highway': [
        'residential',
        'service',
        'unclassified',
    ]
})

# select the building types you want to import
# see: https://wiki.openstreetmap.org/wiki/Key:building
buildings = dict({
    'building': [
        'apartments',
        'commercial',
        'detached',
        'house',
        'industrial',
        'residential',
        'retail',
        'semidetached_house'
    ]
})

# Define a bounding box polygon from a list of lat/lon coordinates
bbox = [(9.1008896, 54.1954005),
        (9.1048374, 54.1961024),
        (9.1090996, 54.1906397),
        (9.1027474, 54.1895923),
        ]
polygon = geometry.Polygon(bbox)
graph = ox.graph_from_polygon(polygon, network_type='drive_service')
ox.plot_graph(graph)

gdf_poly_houses = ox.features_from_polygon(polygon, tags=buildings)
gdf_lines_streets = ox.features_from_polygon(polygon, tags=streets)

# Make sure that only polygon geometries are used
gdf_poly_houses = gdf_poly_houses[gdf_poly_houses['geometry'].apply(
    lambda x: isinstance(x, geometry.Polygon))]

# We need one (or more) buildings that we call "generators".
# Choose one among the buildings at random and move it to a new GeoDataFrame
np.random.seed(42)
id_generator = np.random.randint(len(gdf_poly_houses))
gdf_poly_gen = gdf_poly_houses.iloc[[id_generator]].copy()
gdf_poly_houses.drop(index=gdf_poly_houses.index[id_generator], inplace=True)

# The houses need a maximum thermal power. For this example, we set it
# to a random value between 10 and 50 kW for all houses
gdf_poly_houses['P_heat_max'] = \
    np.random.randint(10, 50, size=len(gdf_poly_houses))

# plot the given geometry
fig, ax = plt.subplots()
gdf_lines_streets.plot(ax=ax, color='blue')
gdf_poly_gen.plot(ax=ax, color='orange')
gdf_poly_houses.plot(ax=ax, color='green')
plt.title('Geometry before processing')
plt.show()

# # Optionally export the imported geometry (e.g. for QGIS)
# gdf_poly_houses.to_file('footprint_buildings.geojson', driver='GeoJSON')

# Part II: Process the geometry for DHNx #############

# # optionally you can skip Part I and load your own layer with geopandas, e.g.
# gdf_lines_streets = gpd.read_file('your_file.geojson')
# gdf_poly_gen = gpd.read_file('your_file.geojson')
# gdf_poly_houses = gpd.read_file('your_file.geojson')

# process the geometry
tn_input = process_geometry(
    lines=gdf_lines_streets,
    producers=gdf_poly_gen,
    consumers=gdf_poly_houses,
    method="boundary",  # select the method of how to connect the buildings
)

# plot output after processing the geometry
_, ax = plt.subplots()
tn_input['consumers'].plot(ax=ax, color='green')
tn_input['producers'].plot(ax=ax, color='red')
tn_input['pipes'].plot(ax=ax, color='blue')
tn_input['forks'].plot(ax=ax, color='grey')
plt.title('Geometry after processing')
plt.show()

# # optionally export the geodataframes and load it into qgis, arcgis whatever
# # for checking the results of the geometry processing
# path_geo = 'qgis'
# for key, val in tn_input.items():
#     val.to_file(os.path.join(path_geo, key + '.geojson'), driver='GeoJSON')


# Part III: Initialise the ThermalNetwork and perform the Optimisation #######

# initialize a ThermalNetwork
network = ThermalNetwork()

# add the pipes, forks, consumer, and producers to the ThermalNetwork
for k, v in tn_input.items():
    network.components[k] = v

# check if ThermalNetwork is consistent
network.is_consistent()

# load the specification of the oemof-solph components
invest_opt = load_invest_options('./invest_data')


# optionally, define some settings for the solver. Especially increasing the
# solution tolerance with 'ratioGap' or setting a maximum runtime in 'seconds'
# helps if large networks take too long to solve
settings = dict(solver='cbc',
                solve_kw={
                    'tee': False,  # print solver output
                },
                solver_cmdline_options={
                    # 'allowableGap': 1e-5,  # (absolute gap) default: 1e-10
                    # 'ratioGap': 0.2,  # (0.2 = 20% gap) default: 0
                    # 'seconds': 60 * 1,  # (maximum runtime) default: 1e+100
                },
                )

# perform the investment optimisation
network.optimize_investment(invest_options=invest_opt, **settings)


# Part IV: Check the results #############

# get results
results_edges = network.results.optimization['components']['pipes']
# print(results_edges[['from_node', 'to_node', 'hp_type', 'capacity',
#                      'direction', 'costs', 'losses']])

print(results_edges[['costs']].sum())
print('Objective value: ', network.results.optimization['oemof_meta']['objective'])
# (The costs of the objective value and the investment costs of the DHS
# pipelines are the same, since no additional costs (e.g. for energy sources)
# are considered in this example.)

# add the investment results to the geoDataFrame
gdf_pipes = network.components['pipes']
gdf_pipes = gdf_pipes.join(results_edges, rsuffix='results_')

# plot output after processing the geometry
_, ax = plt.subplots()
network.components['consumers'].plot(ax=ax, color='green')
network.components['producers'].plot(ax=ax, color='red')
network.components['forks'].plot(ax=ax, color='grey')
gdf_pipes[gdf_pipes['capacity'] > 0].plot(ax=ax, color='blue')
plt.title('Invested pipelines')
plt.show()



# -*- coding: utf-8 -*-
"""
These tests test little examples for the use of the library.

This file is part of project oemof (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/tools/helpers.py

SPDX-License-Identifier: MIT
"""

import os

import dhnx

basedir = os.path.dirname(__file__)

dir_import = os.path.join(basedir, '_files/looped_network_import')

thermal_network = dhnx.network.ThermalNetwork(dir_import)


def test_add():
    thermal_network.add('Producer', 5, lat=1, lon=1)

    assert thermal_network.components['producers'].loc[5].to_list() == [1., 1., "Producer"]


def test_remove():
    thermal_network.remove('Consumer', 1)

    assert 4 not in thermal_network.components['consumers'].index



# -*- coding: utf-8

"""
This module is designed to hold optimization model implementations. The
implementation makes use of oemof-solph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

import logging
import os

import networkx as nx
import oemof.solph as solph
import pandas as pd
from oemof.solph import helpers

from dhnx.model import InvestOptimizationModel
from dhnx.model import OperationOptimizationModel
from dhnx.optimization.dhs_nodes import add_nodes_dhs
from dhnx.optimization.dhs_nodes import add_nodes_houses

logger = logging.getLogger(__name__)  # Create a logger for this module


class OemofOperationOptimizationModel(OperationOptimizationModel):
    r"""
    Implementation of an operation optimization model using oemof-solph.
    """
    def __init__(self, thermal_network):
        super().__init__(thermal_network)
        self.results = {}

    def setup(self):
        pass

    def solve(self):
        pass

    def get_results(self):
        return self.results


class OemofInvestOptimizationModel(InvestOptimizationModel):
    """
    Implementation of an invest optimization model using oemof-solph.

    ...

    Attributes
    ----------
    settings : dict
        Dictionary holding the optimisation settings. See .
    invest_options : dict
        Dictionary holding the investment options for the district heating system.
    nodes : list
        Empty list for collecting all oemof.solph nodes.
    buses : dict
        Empty dictionary for collecting all oemof.solph.Buses of the energy system.
    es : oemof.solph.EnergySystem
        Empty oemof.solph.EnergySystem.
    om : oemof.solph.Model
        Attribute, which will be the oemof.solph.Model for optimisation.
    oemof_flow_attr : set
        Possible flow attributes, which can be used additionally:
        {'nominal_capacity', 'min', 'max', 'variable_costs', 'fix'}
    results : dict
        Empty dictionary for the results.

    Methods
    -------
    check_input():
        Performs checks on the input data.
    complete_exist_data():
        Sets the investment status for the results dataframe of the pipes.
    get_pipe_data():
        Adds heat loss and investment costs to pipes dataframe.
    setup_oemof_es():
        The energy system *es* is build.
    setup():
        Calls *check_input()*, *complete_exist_data()*, *get_pipe_data()*, and *setup_oemof_es()*.

    """
    def __init__(self, thermal_network, settings, investment_options):

        self.settings = settings
        self.invest_options = investment_options
        self.nodes = []  # list of all nodes
        self.buses = {}  # dict of all buses
        self.es = solph.EnergySystem()
        self.om = None

        # list of possible oemof flow attributes, e.g. for producers source
        self.oemof_flow_attr = {'nominal_capacity', 'min', 'max',
                                'variable_costs', 'fix'}

        super().__init__(thermal_network)
        self.results = {}

    def check_input(self):
        """Check 1:

        Check and make sure, that the dtypes of the columns of the sequences
        and the indices (=ids) of the forks, pipes, producers and consumers
        are of type 'str'. (They need to be the same dtye.)

        Check 2:

        Firstly, it is checked, if there are any not-allowed connection in the *pipe* data.
        The following connections are not allowed:

          * consumer -> consumer
          * producer -> producer
          * producer -> consumer
          * consumer -> fork

        Secondly, it is checked, if a pipes goes to a consumer, which does not exist.

        Check 3

        Checks if graph of network is connected.

        An error is raised if one of these connection occurs.
        """

        # Check 1
        # make sure that all ids are of type str
        # sequences
        sequ_items = self.thermal_network.sequences.keys()
        for it in sequ_items:
            for v in self.thermal_network.sequences[it].values():
                v.columns.astype('str')

        # components
        for comp in ['pipes', 'consumers', 'producers', 'forks']:
            self.thermal_network.components[comp].index = \
                self.thermal_network.components[comp].index.astype('str')

        # Check 2

        ids_consumers = self.thermal_network.components['consumers'].index

        for p, q in self.thermal_network.components['pipes'].iterrows():

            if (q['from_node'].split('-', 1)[0] == "consumers") and (
                    q['to_node'].split('-', 1)[0] == "consumers"):

                raise ValueError(
                    ""
                    "Pipe id {} goes from consumer to consumer. This is not "
                    "allowed!".format(p))

            if (q['from_node'].split('-', 1)[0] == "producers") and (
                    q['to_node'].split('-', 1)[0] == "producers"):

                raise ValueError(
                    ""
                    "Pipe id {} goes from producers to producers. "
                    "This is not allowed!".format(p))

            if ((q['from_node'].split('-', 1)[0] == "producers") and (
                    q['to_node'].split('-', 1)[0] == "consumers")) or ((
                        q['from_node'].split('-', 1)[0] == "consumers") and (
                            q['to_node'].split('-', 1)[0] == "producers")):

                raise ValueError(
                    ""
                    "Pipe id {} goes from producers directly "
                    "to consumers, or vice versa. This is not allowed!"
                    "".format(p))

            if (q['from_node'].split('-', 1)[0] == "forks") and (
                    q['to_node'].split('-', 1)[0] == "consumers"):

                cons_id = q['to_node'].split('-', 1)[1]

                if cons_id not in ids_consumers:
                    raise ValueError(
                        ""
                        "The consumer {} of pipe id {} does not exist!"
                        .format(cons_id, p))

        pipe_to_cons_ids = list(self.thermal_network.components['pipes']['to_node'].values)
        pipe_to_cons_ids = [x.split('-', 1)[1] for x in pipe_to_cons_ids
                            if x.split('-', 1)[0] == 'consumers']

        for id in list(self.thermal_network.components['consumers'].index):
            if id not in pipe_to_cons_ids:
                raise ValueError(
                    "The consumer id {} has no connection the the grid!".format(id))

        # Check 3
        # check if all components of network are connected
        self.thermal_network.nx_graph = self.thermal_network.to_nx_graph()
        g = nx.Graph(self.thermal_network.nx_graph)
        if not nx.is_connected(g):
            nx_sum = [len(c) for c in sorted(nx.connected_components(g), key=len, reverse=True)]
            nx_detail = sorted(nx.connected_components(g), key=len, reverse=True)
            raise ValueError(
                "Network not connected! There are {} parts, with the following number of nodes: \n"
                "{} \n"
                "These are the separated elements/networks: \n"
                "{}".format(len(nx_sum), nx_sum, nx_detail)
            )

    def remove_inactive(self):
        """
        If the attribute active is present in any of the components
        columns, or in any the investment options tables,
        all rows with active == 0 are deleted, and the column active
        is deleted.
        """
        def clean_df(df):
            if 'active' in df.columns:
                v_new = df[df['active'] == 1].copy()
                v_new.drop('active', axis=1, inplace=True)
                df = v_new
            return df

        for k, v in self.thermal_network.components.items():
            self.thermal_network.components[k] = clean_df(v)

        pipes = self.invest_options['network']['pipes']
        self.invest_options['network']['pipes'] = clean_df(pipes)

        for node_typ in ['consumers', 'producers']:
            for k, v in self.invest_options[node_typ].items():
                self.invest_options[node_typ][k] = clean_df(v)

    def prepare_heat_demand(self):
        """
        This method performs the pre-processing of the heat demand data, depending on
        the given optimisation settings.

        - If attribute 'P_heat_max' not given at the consumers, the maximum heat demand
          is calculated from the timeseries and added the consumers table.
        - If the optimisation setting 'heat_demand' == scalar, the number of time steps
          of the optimisation is set to 1, and the 'P_heat_max' values are copied to the
          consumers heat flow sequences (which is always the input for the optimisation model).
        - The consumers heat flow sequences are multiplied by the simultaneity factor.
        - Finally, a sufficient length of the heat demand timeseries is checked.

        Returns
        -------
        Updated `.network.components['consumers']` and
        `.network.sequences['consumers']['heat_flow']`
        """
        def check_len_timeseries():
            """
            Check, if given number of timesteps of optimization exceeds the length
            of the given heat demand timeseries.
            """
            if self.settings['num_ts'] > \
                    len(self.thermal_network.sequences['consumers']['heat_flow'].index):
                raise ValueError(
                    'The length of the heat demand timeseries is not sufficient '
                    'for the given number of {} timesteps.'.format(
                        self.settings['num_ts']))

        # prepare heat data, whether global simultanity or timeseries
        if 'P_heat_max' not in list(
                self.thermal_network.components['consumers'].columns):
            df_max = self.thermal_network.sequences['consumers']['heat_flow'].max(). \
                to_frame(name='P_heat_max')

            self.thermal_network.components['consumers'] = \
                self.thermal_network.components['consumers'].join(df_max)

        # check, which optimization type should be performed
        if self.settings['heat_demand'] == 'scalar':
            # just single timestep optimization, overwrite previous!
            self.settings['num_ts'] = 1

            # new approach
            p_max = self.thermal_network.components['consumers']['P_heat_max']
            df_ts = pd.DataFrame(data=[p_max.values],
                                 columns=list(p_max.index),
                                 index=pd.Index([0], name='timestep'))

            # heat load is maximum heat load
            self.thermal_network.sequences['consumers']['heat_flow'] = df_ts

        # apply global simultaneity for demand series
        self.thermal_network.sequences['consumers']['heat_flow'] = \
            self.thermal_network.sequences['consumers']['heat_flow'] * \
            self.settings['simultaneity']

        check_len_timeseries()

    def check_existing(self):
        """
        Checks if the attributes `existing` and `hp_type` are given in the `pipes` table.
        If not, the attribute is added, and set to `None` / 0.

        Checks for all existing pipes, if the heatpipe type is given in the pipe type table
        `.invest_options['network']['pipes']`, and if the capacity is greater than zero.
        """

        # check whether there the 'existing' attribute is present at the pipes
        if 'existing' not in self.thermal_network.components['pipes'].columns:
            self.thermal_network.components['pipes']['existing'] = 0

        # create pipes attribute hp_type, if not in the table so far
        if 'hp_type' not in list(self.thermal_network.components['pipes'].columns):
            self.thermal_network.components['pipes']['hp_type'] = None

        edges = self.thermal_network.components['pipes']
        pipe_types = self.invest_options['network']['pipes']

        hp_list = list({x for x in edges['hp_type'].tolist()
                        if isinstance(x, str)})

        for hp in hp_list:
            if hp not in list(pipe_types['label_3']):
                raise ValueError(
                    "Existing heatpipe type {} is not in the list of "
                    "ACTIVE heatpipe investment options!".format(hp)
                )

        for r, c in edges[edges['existing'] == 1].iterrows():
            if c['capacity'] <= 0:
                raise ValueError(
                    "The `capacity` of the existing pipe with id {} must be greater than 0!"
                    "".format(r)
                )

    def setup_oemof_es(self):
        """The oemof solph energy system is initialised based on the settings,
         and filled with oemof-solph object:

         The oemof-solph objects of the *consumers* and *producers* are defined at the consumers
         and producers investment options.

         For the heating infrastructure, there is a *oemof.solph.Bus* added for every fork,
         and a pipe component for every pipe as defined in */network/pipes.csv*.
         """

        date_time_index = pd.date_range(self.settings['start_date'],
                                        periods=self.settings['num_ts'],
                                        freq=self.settings['frequence'])

        logger.info('Initialize the energy system')

        self.es = solph.EnergySystem(
            timeindex=date_time_index,
            infer_last_interval=True,
        )

        logger.info('Create oemof objects')

        # add houses and generation
        for typ in ['consumers', 'producers']:
            self.nodes, self.buses = add_nodes_houses(
                self, self.nodes, self.buses, typ)

        logger.info('Producers, Consumers Nodes appended.')

        # add heating infrastructure
        self.nodes, self.buses = add_nodes_dhs(self, self.settings, self.nodes,
                                               self.buses)
        logger.info('DHS Nodes appended.')

        # add nodes and flows to energy system
        self.es.add(*self.nodes)

        logger.info('Energysystem has been created')

        if self.settings['print_logging_info']:
            print("*********************************************************")
            print("The following objects have been created:")
            for n in self.es.nodes:
                oobj = \
                    str(type(n)).replace("<class 'oemof.solph.", "").replace("'>",
                                                                             "")
                print(oobj + ':', n.label)
            print("*********************************************************")

    def setup(self):
        """
        Calls *remove_inactive()* *check_input()*, *prepare_heat_demand()*,
        *complete_exist_data()*, and *setup_oemof_es()*.
        """

        # removes all rows with attribute active == 0 - if 'active given
        self.remove_inactive()

        # initial check of pipes connections
        self.check_input()

        # pre-processes the heat demand data depending on optimisation settings
        self.prepare_heat_demand()

        # check if existing pipes are given
        self.check_existing()

        # set up oemof energy system
        self.setup_oemof_es()

    def solve(self):
        """Builds the oemof.solph.Model of the energysystem *es*."""

        logger.info('Build the operational model')
        self.om = solph.Model(self.es)

        if self.settings['solve_kw'] is None:
            s_kw = {"tee": True}
        else:
            s_kw = self.settings['solve_kw']

        logger.info('Solve the optimization problem')
        self.om.solve(
            solver=self.settings['solver'],
            solve_kwargs=s_kw,
            cmdline_options=self.settings.get('solver_cmdline_options', {}))

        if self.settings['write_lp_file']:
            filename = os.path.join(
                helpers.extend_basic_path('lp_files'), 'DHNx.lp')
            logger.info('Store lp-file in %s', filename)
            self.om.write(filename, io_options={'symbolic_solver_labels': True})

        self.es.results['main'] = solph.processing.results(self.om)
        self.es.results['meta'] = solph.processing.meta_results(self.om)

    def get_results_edges(self):
        """Postprocessing of the investment results of the pipes."""

        def get_invest_val(lab):

            res = self.es.results['main']

            outflow = [x for x in res.keys()
                       if x[1] is not None
                       if lab == str(x[0].label)]

            if len(outflow) > 1:
                logger.info('Multiple IDs!')

            try:
                invest = res[outflow[0]]['scalars']['invest']
            except (KeyError, IndexError):
                try:
                    # that's in case of a one timestep optimisation due to
                    # an oemof bug in outputlib
                    invest = res[outflow[0]]['sequences']['invest'].iloc[0]
                except (KeyError, IndexError):
                    # this is in case there is no bi-directional heatpipe,
                    # e.g. at forks-consumers, producers-forks
                    invest = 0

            # the rounding is performed due to numerical issues
            return round(invest, 6)

        def get_invest_status(lab):

            res = self.es.results['main']

            outflow = [x for x in res.keys()
                       if x[1] is not None
                       if lab == str(x[0].label)]

            try:
                invest_status = res[outflow[0]]['scalars']['invest_status']
            except (KeyError, IndexError):
                try:
                    # that's in case of a one timestep optimisation due to
                    # an oemof bug in outputlib
                    invest_status = res[outflow[0]]['sequences']['invest_status'].iloc[0]
                except (KeyError, IndexError):
                    # this is in case there is no bi-directional heatpipe,
                    # e.g. at forks-consumers, producers-forks
                    invest_status = 0

            return invest_status

        def get_hp_results(p):

            hp_lab = p['label_3']
            label_base = 'infrastructure_' + 'heat_' + hp_lab + '_'

            # maybe slow approach with lambda function
            df[hp_lab + '.' + 'dir-1'] = df['from_node'] + '-' + df['to_node']
            df[hp_lab + '.' + 'size-1'] = df[hp_lab + '.' + 'dir-1'].apply(
                lambda x: get_invest_val(label_base + x))
            df[hp_lab + '.' + 'dir-2'] = df['to_node'] + '-' + df['from_node']
            df[hp_lab + '.' + 'size-2'] = df[hp_lab + '.' + 'dir-2'].apply(
                lambda x: get_invest_val(label_base + x))

            df[hp_lab + '.' + 'size'] = \
                df[[hp_lab + '.' + 'size-1', hp_lab + '.' + 'size-2']].max(axis=1)

            # get direction of pipes
            for r, c in df.iterrows():
                if c[hp_lab + '.' + 'size-1'] > c[hp_lab + '.' + 'size-2']:
                    df.at[r, hp_lab + '.direction'] = 1
                elif c[hp_lab + '.' + 'size-1'] < c[hp_lab + '.' + 'size-2']:
                    df.at[r, hp_lab + '.direction'] = -1
                else:
                    df.at[r, hp_lab + '.direction'] = 0

            if p['nonconvex']:
                df[hp_lab + '.' + 'status-1'] = df[hp_lab + '.' + 'dir-1'].apply(
                    lambda x: get_invest_status(label_base + x))
                df[hp_lab + '.' + 'status-2'] = df[hp_lab + '.' + 'dir-2'].apply(
                    lambda x: get_invest_status(label_base + x))
                df[hp_lab + '.' + 'status'] = \
                    df[[hp_lab + '.' + 'status-1', hp_lab + '.' + 'status-2']].max(axis=1)

                for r, c in df.iterrows():
                    if df.at[r, hp_lab + '.' + 'status-1'] + \
                            df.at[r, hp_lab + '.' + 'status-2'] > 1:
                        logger.warning(
                            "Investment status of pipe id {} is 1 for both dircetions!"
                            " This is not allowed!".format(r)
                        )
                    if (df.at[r, hp_lab + '.' + 'status-1'] == 1 and df.at[
                            r, hp_lab + '.' + 'size-1'] == 0) \
                            or\
                            (df.at[r, hp_lab + '.' + 'status-2'] == 1 and df.at[
                                r, hp_lab + '.' + 'size-2'] == 0):
                        logger.warning(
                            "Investment status of pipe id {} is 1, and capacity is 0!"
                            "What happend?!".format(r)
                        )

            return df

        def check_multi_dir_invest(hp_lab):

            df_double_invest = \
                df[(df[hp_lab + '.' + 'size-1'] > 0) & (df[hp_lab + '.' + 'size-2'] > 0)]

            if self.settings['print_logging_info']:
                print('***')
                if df_double_invest.empty:
                    print('There is NO investment in both directions at the'
                          'following pipes for "', hp_lab, '"')
                else:
                    print('There is an investment in both directions at the'
                          'following pipes for "', hp_lab, '":')
                    print('----------')
                    print(' id | from_node | to_node | size-1 | size-2 ')
                    print('============================================')
                    for r, c in df_double_invest.iterrows():
                        print(r, ' | ', c['from_node'], ' | ', c['to_node'],
                              ' | ', c[hp_lab + '.' + 'size-1'], ' | ',
                              c[hp_lab + '.' + 'size-2'], ' | ')
                    print('----------')

        def catch_up_results():

            def check_invest_label(hp_type, edge_id):
                """
                If there is already a heatpipe type, an error is raised,
                because only one investment type for each edges makes sense.
                """
                if isinstance(hp_type, str):
                    raise ValueError(
                        "Pipe id {} already has an investment > 0!".format(edge_id))

            df['hp_type'] = None
            df['capacity'] = float(0)
            df['direction'] = 0
            df['status'] = float(0)

            for ahp in active_hp:
                # p = df_hp[df_hp['label_3'] == ahp].squeeze()   # series of heatpipe
                for r, c in df.iterrows():
                    if c[ahp + '.size'] > 0:
                        check_invest_label(c['hp_type'], id)
                        df.at[r, 'hp_type'] = ahp
                        df.at[r, 'capacity'] = c[ahp + '.size']
                        df.at[r, 'direction'] = c[ahp + '.direction']
                        if ahp + '.status' in c.index:
                            df.at[r, 'status'] = c[ahp + '.status']

        def recalc_costs_losses():
            """
            Calculates the investment costs and thermal losses for each
            pipeline of the district heating network.
            (This recalculation could also serve as check for comparing the
            total pipeline costs with the objective value of oemof.solph
            optimisation model)

            Note
            ----
            With nonconvex investments (with binary variables) it could happen
            that very low results of the decision variable occur (although
            there is a minimum investment threshold), and that
            the status variable could have values of almost 0 and almost 1.
            This cost re-calculation does not round any of this results, and
            transfers the results as they into a DataFrame containing all
            pipes of the district heating network.

            """

            df['costs'] = float(0)
            df['losses'] = float(0)

            for r, c in df.iterrows():
                if c['capacity'] > 0:
                    hp_lab = c['hp_type']
                    # select row from heatpipe type table
                    hp_p = df_hp[df_hp['label_3'] == hp_lab].squeeze()
                    if hp_p['nonconvex'] == 1:
                        df.at[r, 'costs'] = c['length'] * (
                            c['capacity'] * hp_p['capex_pipes'] +  # noqa: W504
                            hp_p['fix_costs'] * c['status']
                        )
                        df.at[r, 'losses'] = c['length'] * (
                            c['capacity'] * hp_p['l_factor'] +  # noqa: W504
                            hp_p['l_factor_fix'] * c['status']
                        )

                    elif hp_p['nonconvex'] == 0:
                        df.at[r, 'costs'] = c['length'] * c['capacity'] * hp_p['capex_pipes']
                        # Note, that a constant loss is possible also for convex
                        df.at[r, 'losses'] = c['length'] * (
                            c['capacity'] * hp_p['l_factor'] + hp_p['l_factor_fix']
                        )

        # use pipes dataframe as base and add results as new columns to it
        df = self.thermal_network.components['pipes']

        # only select not existing pipes
        df = df[df['existing'] == 0].copy()

        # remove input data
        df = df[['from_node', 'to_node', 'length']].copy()

        # putting the results of the investments in heatpipes to the pipes:
        df_hp = self.invest_options['network']['pipes']

        # list of active heat pipes
        active_hp = list(df_hp['label_3'].values)

        for hp in active_hp:
            hp_param = df_hp[df_hp['label_3'] == hp].squeeze()
            get_hp_results(hp_param)
            check_multi_dir_invest(hp)

        catch_up_results()

        recalc_costs_losses()

        return df[['from_node', 'to_node', 'length', 'hp_type', 'capacity', 'direction',
                   'costs', 'losses']]


def optimize_operation(thermal_network):
    r"""
    Takes a thermal network and returns the result of
    the operational optimization.
    """
    model = OemofOperationOptimizationModel(thermal_network)

    model.solve()

    results = model.get_results()

    return results


def setup_optimise_investment(
        thermal_network, invest_options, heat_demand='scalar', num_ts=1,
        time_res=1, start_date='1/1/2018', frequence='h', solver='cbc',
        solve_kw=None, solver_cmdline_options=None, simultaneity=1,
        bidirectional_pipes=False, dump_path=None, dump_name='dump.oemof',
        print_logging_info=False, write_lp_file=False):
    """
    Function for setting up the oemof solph operational Model.

    Parameters
    ----------
    thermal_network : ThermalNetwork
        See the ThermalNetwork class.
    invest_options : dict
        Dictionary holding the investment options for the district heating system.
    heat_demand : str
        'scalar': Peak heat load is used as heat consumers’ heat demand.
        'series': Heat load time-series is used.
    num_ts : int
        Number of time steps of optimisation.
    time_res : float
        Time resolution.
    start_date : str or datetime-like
        Startdate for oemof optimisation.
    frequence : str or DateOffset
        Lenght of period.
    solver : str
        Name of solver.
    solve_kw : dict
        Solver kwargs.
    solver_cmdline_options : dict
        Dictionary with command line options for solver.
    simultaneity : float
        Simultaneity factor.
    bidirectional_pipes : bool
        Bidirectional pipes leads to bi-directional flow attributes at the heatpipeline components
        {‘min’: -1, bidirectional: True}.
    dump_path : str
        If a dump path is provided, the oemof dump file is stored.
    dump_name : str
        Name of dump file.
    print_logging_info : bool
        Additional logging info is printed.
    write_lp_file : bool
        Linear program file is stored (‘User/.oemof/lp_files/DHNx.lp’).

    Returns
    -------
    oemof.solph.Model : The oemof.solph.Model is build.
    """
    if heat_demand not in ['scalar', 'series']:
        raise ValueError(
            'The settings attribute *heat_demand* must be "scalar" or "series"!'
        )

    if solver_cmdline_options is None:
        solver_cmdline_options = {}

    settings = {
        'heat_demand': heat_demand,
        'num_ts': num_ts,
        'time_res': time_res,
        'start_date': start_date,
        'frequence': frequence,
        'solver': solver,
        'solve_kw': solve_kw,
        'solver_cmdline_options': solver_cmdline_options,
        'simultaneity': simultaneity,
        'bidirectional_pipes': bidirectional_pipes,
        'dump_path': dump_path,
        'dump_name': dump_name,
        'print_logging_info': print_logging_info,
        'write_lp_file': write_lp_file,
    }

    model = OemofInvestOptimizationModel(thermal_network, settings, invest_options)

    return model


def solve_optimisation_investment(model):
    """

    Parameters
    ----------
    model : oemof.solph.Model
        The oemof model, which is optimized.

    Returns
    -------
    dict : Results of optimisation. Contains:
        - 'oemof' : Complete "oemof" results of the energy system optimisation (.results['main']).
        - 'oemof_meta' : Meta results of oemof solph optimisation.
        - 'components' : 'pipes' : Investment results of pipes.
    """

    model.solve()

    if model.settings['dump_path'] is not None:
        my_es = model.es
        my_es.dump(dpath=model.settings['dump_path'], filename=model.settings['dump_name'])
        logger.info('oemof Energysystem stored in "{}"'.format(model.settings['dump_path']))

    edges_results = model.get_results_edges()

    results = {'oemof': model.es.results['main'],
               'oemof_meta': model.es.results['meta'],
               'components': {'pipes': edges_results}}

    return results



import matplotlib.pyplot as plt

from dhnx.input_output import OSMNetworkImporter
from dhnx.plotting import StaticMap
from dhnx.network import ThermalNetwork

place = (52.43034, 13.53806)

distance = 300

tnw = ThermalNetwork()

io = OSMNetworkImporter(tnw, place, distance)

io.load()

print(tnw)

# plot static map
static_map = StaticMap(tnw)

static_map.draw(background_map=False)

plt.savefig('static_map_wo_background.png')

tnw.to_csv_folder('osm_network')



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from dhnx.optimization.precalc_hydraulic import v_max_bisection, \
    calc_mass_flow, calc_power, v_max_secant

df = pd.read_csv("Pipe_data.csv", sep=";")

maximum_pressure_drop = 150  # in Pa/m

# Calculation of maximum velocity
df['v_max [m/s]'] = df.apply(lambda row: v_max_bisection(
    d_i=row['Inner diameter [m]'],
    T_average=row['Temperature level [Celsius]'],
    k=row['Roughness [mm]'],
    p_max=maximum_pressure_drop), axis=1)

# alternative with secant method
df['v_max (secant) [m/s]'] = df.apply(lambda row: v_max_secant(
    d_i=row['Inner diameter [m]'],
    T_average=row['Temperature level [Celsius]'],
    k=row['Roughness [mm]'],
    p_max=maximum_pressure_drop), axis=1)

# Calculation of mass flow
df['Mass flow [kg/s]'] = df.apply(lambda row: calc_mass_flow(
    v=row['v_max [m/s]'], di=row['Inner diameter [m]'],
    T_av=row['Temperature level [Celsius]']), axis=1)

# Calculation of maximum Power
df['P_max [kW]'] = df.apply(lambda row: 0.001 * calc_power(
    T_vl=row['T_forward [Celsius]'],
    T_rl=row['T_return [Celsius]'],
    mf=row['Mass flow [kg/s]']), axis=1)

# Create pipes table for district heating network optimization

# Linear approximation with 1 segment
constants_costs = np.polyfit(df['P_max [kW]'], df['Costs [eur]'], 1)
constants_loss = np.polyfit(df['P_max [kW]'], df['Loss [kW/m]'], 1)

print('Costs constants: ', constants_costs)
print('Loss constants: ', constants_loss)

df_pipes = pd.DataFrame(
    {
        "label_3": "your-pipe-type-label",
        "active": 1,
        "nonconvex": 1,
        "l_factor": constants_loss[0],
        "l_factor_fix": constants_loss[1],
        "cap_max": df['P_max [kW]'].max(),
        "cap_min": df['P_max [kW]'].min(),
        "capex_pipes": constants_costs[0],
        "fix_costs": constants_costs[1],
    }, index=[0],
)

x_min = df['P_max [kW]'].min()
x_max = df['P_max [kW]'].max()
y_min = constants_costs[0] * x_min + constants_costs[1]
y_max = constants_costs[0] * x_max + constants_costs[1]

fig, ax = plt.subplots()
x = df['P_max [kW]']
y = df['Costs [eur]']
ax.plot(x, y, lw=0, marker="o")
ax.plot(
    [x_min, x_max], [y_min, y_max],
    ls=":", color='r', marker="x"
)
ax.set_xlabel("Transport capacity [kW]")
ax.set_ylabel("Kosten [€/m]")
plt.text(2000, 250, "Linear cost approximation \n"
                    "of district heating pipelines \n"
                    "based on maximum pressure drop \n"
                    "of {:.0f} Pa/m".format(maximum_pressure_drop))
plt.ylim(0, None)
plt.grid(ls=":")
plt.show()

# TODO : Show linear approximation with more segments



# -*- coding: utf-8

"""
This module is designed to hold functions for pre-calculation of hydraulic
parameters for the district heating network dimensioning.

The aim is to calculate the maximum heat transport capacity
of pipelines given a maximum pressure drop per meter, roughness of the pipes
inner surface, and an estimated delta T of the forward and return pipes.

The equations and values used for the calculation can be found here:
http://www.math-tech.at/Beispiele/upload/gra_Druckverlust_in_Rohrleitungen.PDF
https://www.schweizer-fn.de/stroemung/rauhigkeit/rauhigkeit.php

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""

import logging
import math

import numpy as np
from scipy.optimize import fsolve

try:
    from CoolProp.CoolProp import PropsSI

except ImportError:
    print("Need to install CoolProp to use the hydraulic "
          "pre-calculation module.")

logger = logging.getLogger(__name__)  # Create a logger for this module


def eq_smooth(x, R_e):
    r"""
    Calculation of the pressure drop of hydraulic smooth surfaces.
    (Prandtl & Karman)

    .. eq_smooth_equation:

    :math:`f(x) = x-2 \cdot log\Big(\frac{Re}{2,51x}\Big)`

    Parameters
    ----------
    x: numeric
        :math:`x`: function variable [-]
    R_e: numeric
        :math:`Re`: Reynolds number [-]

    Returns
    -------
    Equation : numeric

    """
    return x - 2 * np.log10(R_e / (x * 2.51))


def eq_transition(x, R_e, k, d_i):
    r"""
    Equation to be solved for the transition range
    between a smooth and rough pipe surface (Prandtl-Colebrook)

    .. eq_transition_equation

    :math:`f(x)=x+2\cdot log \big(  \frac{2,51x}{Re} \cdot \frac{k}{3,71d_i} \big)`

    Parameters
    ----------
    x : numeric
        :math:`x`: function variable () [-]

    R_e: numeric
        :math:`Re`: Reynolds number [-]

    k : numeric
        :math:`k`: roughness of inner pipeline surface [mm]

    d_i : numeric
        :math:`d_i`: inner diameter [m]

    Returns
    -------
    Equation : numeric

    """

    return x + 2 * np.log10((2.51 * x) / R_e + k / (3.71 * d_i))


def calc_k_v(d_v, d):
    r"""
    Calcutlates the kinematic viscosity for given density and dynamic viscosity

    .. calc_k_v_equation:

    :math:`\nu = \frac{\eta}{\rho}`

    Parameters
    ----------
    d_v: numeric
        :math:`\eta`: dynamic viskosity [kg/(m*s)]

    d: numeric
        :math:`\rho`: density [kg/m³]

    Returns
    -------
    kinematic viscosity [m²/s] : numeric

    """
    return d_v / d


def calc_Re(v, d_i, k_v):
    r"""
    Calculates the Reynolds number for a given velocity, inner diameter
    and kinematic viscosity

    .. calc_Re_equation:

    :math:`\frac{v \cdot d_i}{\nu}`

    Parameters
    ----------
    v: numeric
        :math:`v`: flow velocity [m/s]

    d_i: numeric
        :math:`d_i`: inner pipe diameter [m]

    k_v: numeric
        :math:`\nu`: kinematic viscosity [m²/s]

    Returns
    -------
    Reynolds number [-]: numeric

    """
    return v * d_i / k_v


def calc_lambda_laminar(Re):
    r"""
    Calculates the Darcy friction factor for a given Reynolds number
    for a laminar flow

    .. calc_lam_lam_equation:

    :math:`\lambda=\frac{64}{Re}`

    Parameters
    ----------
    Re: numeric
        :math:`Re`: Reynolds number [-]

    Returns
    -------
    Darcy friction factor [-] : numeric

    """
    return 64 / Re


def calc_d_p(lam, length, d_i, d, v):
    r"""
    Calculates the pressure drop in a pipe for a given Darcy friction factor

    .. calc_d_p_equation:

    :math:`\Delta p = \lambda \frac{l}{d_i} \frac{\rho}{2} v^2`

    Parameters
    ----------
    lam: numeric
        :math:`\lambda`: Darcy friction factor [-]

    length: numeric
        :math:`l`: length of the pipe [m]

    d_i : numeric
        :math:`d_i`: inner pipe diameter [m]

    d: numeric
        :math:`\rho`: density [kg/m³]

    v: numeric
        :math:`v`: flow velocity [m/s]

    Returns
    -------
    Pressure drop [Pa]: numeric
    """
    return lam * length / d_i * d / 2 * v ** 2


def calc_lambda_turb1(Re):
    r"""
    Calculates the Darcy friction factor for a given Reynolds number
    for a turbulent flow, a smooth pipe and a Reynolds number smaller
    than 10^5 (Blasius)

    .. calc_lam_turb1_equation

    :math:`\lambda = 0,3164\cdot Re ^{-0,25}`

    Parameters
    ----------
    Re: numeric
        :math:`Re`: Reynolds number [-]

    Returns
    -------
    Darcy friction factor [-] : numeric

    """
    return 0.3164 * Re ** (-0.25)


def calc_lambda_turb2(Re):
    r"""
    Calculates the Darcy friction factor for a given Reynolds number
    for a turbulent flow, a smooth pipe and a Reynolds number
    between 10^5 and 10^6 (Nikuradse)

    .. calc_lam_turb1_equation

    :math:`\lambda = 0,0032 + 0,221 \cdot Re ^{-0,237}`

    Parameters
    ----------
    Re: numeric
        :math:`Re`: Reynolds number [-]

    Returns
    -------
    Darcy friction factor [-] : numeric

    """
    return 0.0032 + 0.221 * Re ** (-0.237)


def calc_lambda_turb3(Re):
    r"""
    Calculates the Darcy friction factor for a given Reynolds number
    for a turbulent flow, a smooth pipe and a Reynolds number higher than 10^6.
    For a formula, see :func:`~precalc_hydraulic.eq_smooth`.

    Parameters
    ----------
    Re: numeric
        :math:`Re`: Reynolds number [-]

    Returns
    -------
    Darcy friction factor [-] : numeric

    """
    lam_init = 0.3164 / (Re ** 0.25)

    x = fsolve(eq_smooth, x0=lam_init, args=Re)

    return 1 / x[0] ** 2


def calc_lambda_rough(d_i, k):
    r"""
    Calculates the Darcy friction factor for a turbulent flow
    and a rough inner pipe surface (Prandtl & Nikuradse)

    .. calc_lam_rough_equation

    :math:`\lambda =\frac{1}{(2 log(3,71\frac{d_i}{k}))^2}`

    Parameters
    ----------
    d_i : numeric
        :math:`d_i`: inner pipe diameter [m]

    k : numeric
        :math:`k`: roughness of inner pipeline surface [mm]

    Returns
    -------
    Darcy friction factor [-] : numeric

    """
    return 1 / ((-2 * np.log10(k / (3.71 * d_i))) ** 2)


def calc_lambda_transition(R_e, k, d_i):
    r"""
    Calculates the Darcy friction factor for a given Reynolds number
    for a turbulent flow and the transition area between a rough and
    smooth pipe surface.

    See also :func:`~precalc_hydraulic.eq_transition`.

    Parameters
    ----------
    R_e: numeric
        :math:`Re`: Reynolds number [-]

    k : numeric
        :math:`k`: roughness of inner pipeline surface [mm]

    d_i : numeric
        :math:`d_i`: inner pipe diameter [m]

    Returns
    -------
    Darcy friction factor [-] : numeric

    """
    lam_init = 0.25 / R_e ** 0.2
    x = fsolve(eq_transition, x0=lam_init, args=(R_e, k, d_i))
    return 1 / x[0] ** 2


def delta_p(v, d_i, k=0.1, T_medium=90, length=1,
            pressure=101325, R_crit=2320, fluid='IF97::Water'):

    r"""
    Function to calculate the pressure loss in a pipeline

    Parameters
    ----------
    v : numeric
        :math:`v`: flow velocity [m/s]

    d_i : numeric
        :math:`d_i`: inner pipe diameter [m]

    k : numeric
        :math:`k`: roughness of inner pipeline surface [mm]

    T_medium : numeric
        :math:`T_{medium}`: fluid temperature [°C]

    length : numeric
        :math:`l`: length of the pipe [m]

    pressure : numeric
        :math:`p`: pressure in the pipe [Pa]

    R_crit : numeric
        :math:`Re_{crit}`: critical Reynolds number between laminar and turbulent flow [-]

    fluid : str
        name of the fluid used

    Returns
    -------
    Pressure drop [Pa] : numeric

    """
    k = k * 0.001

    # get density of water [kg/m^3]
    d = PropsSI('D', 'T', T_medium + 273.15, 'P', pressure, fluid)
    # dynamic viscosity eta [kg/(m*s)]
    d_v = PropsSI('V', 'T', T_medium + 273.15, 'P', pressure, fluid)
    k_v = calc_k_v(d_v, d)

    # Reynolds number
    R_e = calc_Re(v, d_i, k_v)

    if R_e < R_crit:  # laminar flow
        lam = calc_lambda_laminar(R_e)
        d_p = calc_d_p(lam, length, d_i, d, v)

    else:  # turbulent flow

        if R_e * k / d_i < 65:
            # Smooth pipe

            if R_e < 10**5:
                lam = calc_lambda_turb1(R_e)

            elif R_e >= 10**5 and R_e < 10**6:
                lam = calc_lambda_turb2(R_e)

            else:
                # Re > 10^6
                lam = calc_lambda_turb3(R_e)

        elif R_e * k / d_i > 1300:
            # Rough pipe
            lam = calc_lambda_rough(d_i, k)

        else:
            # Transition range 65 < Re * k/d < 1300
            lam = calc_lambda_transition(R_e, k, d_i)

        d_p = calc_d_p(lam, length, d_i, d, v)

    return d_p


def calc_v(vol_flow, d_i):
    r"""
    Calculates the velocity for a given volume flow and inner diameter of a pipe.

    .. calc_v_equation:

    :math:`v_{flow}=\frac{\dot{V}}{(\frac{d_i}{2})^2*\pi}`

    Parameters
    ----------
    vol_flow: numeric
        :math:`\dot{V}`: volume flow [m³/h]

    d_i: numeric
        :math:`d_i`: inner diameter [m]

    Returns
    -------
    flow velocity [m/s] : numeric

    """
    return vol_flow / ((d_i * 0.5)**2 * math.pi * 3600)


def v_max_secant(d_i, T_average, k=0.1, p_max=100, p_epsilon=1,
                 v_0=1, v_1=2,
                 pressure=101325, fluid='IF97::Water'):
    r"""Calculates the maximum velocity via iterative approach
    using the secant method.

    The two different starting values v_0 and v_1 should be in the
    area of the maximum flow velocity, as iteration starts from there.

    Parameters
    ----------
    d_i: numeric
        :math:`d_i`: inner diameter [m]

    T_average: numeric
        :math:`T_{av}`: average temperature [°C]

    k: numeric
        :math:`k`: roughness of inner pipeline surface [mm]

    p_max: numeric
        :math:`p_{max}`: maximum pressure drop in pipeline [Pa]

    p_epsilon: numeric
        :math:`p_\epsilon`: accuracy of pressure [Pa]

    v_0: numeric
        :math:`v_0`: first value of initial guess for maximum flow velocity [m/s]

    v_1: numeric
        :math:`v_1`: second value of initial guess for maximum flow velocity [m/s]

    pressure: numeric
        :math:`p`: pressure level [pa]

    fluid: str
        type of fluid, default: 'IF97::Water'

    Returns
    -------
    maximum flow velocity [m/s] : numeric

    """
    p_new = 0
    v_new = 0
    n = 0
    while n < 100:
        n += 1

        p_0 = delta_p(v_0, k=k, d_i=d_i, T_medium=T_average,
                      pressure=pressure, fluid=fluid)

        p_1 = delta_p(v_1, k=k, d_i=d_i, T_medium=T_average,
                      pressure=pressure, fluid=fluid)

        v_new = v_1 - (p_1 - p_max) * (v_1 - v_0) / (p_1 - p_0)

        p_new = delta_p(v_new, k=k, d_i=d_i, T_medium=T_average,
                        pressure=pressure, fluid=fluid)

        if abs(p_new - p_max) < p_epsilon:
            break

        else:
            v_0 = v_1
            v_1 = v_new

    logger.info(
        "Maximum flow velocity calculated. Iterations: %d, "
        "Flow velocity: %.4f [m/s], Pressure drop: %.4f [Pa/m]"
        % (n, v_new, p_new)
    )

    return v_new


def v_max_bisection(d_i, T_average, k=0.1, p_max=100,
                    p_epsilon=0.1, v_epsilon=0.001,
                    v_0=0.01, v_1=10,
                    pressure=101325, fluid='IF97::Water'):
    r"""Calculates the maximum velocity via bisection for a given pressure drop.

    The two starting values `v_0` and `v_1` need to be given,
    with `v_0` below the expected flow velocity and `v_1` above.
    These are the starting values for the bi-section method.

    If either of the stop-criteria `p_epsilon` or `v_epsilon` is reached,
    the iterative calculation is stopped.

    Parameters
    ----------
    d_i: numeric
        :math:`d_i`: inner diameter [m]

    T_average: numeric
        :math:`T_{av}`: average temperature [°C]

    k: numeric
        :math:`k`: roughness of inner pipeline surface [mm]

    p_max: numeric
        :math:`p_{max}`: maximum pressure drop in pipeline [Pa]

    p_epsilon: numeric
        :math:`p_\epsilon`: accuracy of pressure [Pa]

    v_epsilon: numeric
        :math:`v_\epsilon`: accuracy of velocity [m/s]

    v_0 : numeric
        :math:`v_0`: first value of initial guess for maximum flow velocity [m/s]

    v_1: numeric
        :math:`v_1`: second value of initial guess for maximum flow velocity [m/s]

    pressure: numeric
        :math:`p`: pressure level [Pa]

    fluid: str
        type of fluid, default: 'IF97::Water'

    Returns
    -------
    maximum flow velocity [m/s] : numeric

    """
    p_0 = delta_p(v_0, k=k, d_i=d_i, T_medium=T_average,
                  pressure=pressure, fluid=fluid)

    p_1 = delta_p(v_1, k=k, d_i=d_i, T_medium=T_average,
                  pressure=pressure, fluid=fluid)

    if (p_0 - p_max) * (p_1 - p_max) >= 0:
        raise AttributeError(
            "The initial guesses `v_0` and `v_1` must be "
            "below and above the expected flow velocity."
        )

    p_new = 0
    v_new = 0
    n = 0
    while n < 200:
        n += 1

        p_0 = delta_p(v_0, k=k, d_i=d_i, T_medium=T_average,
                      pressure=pressure, fluid=fluid)

        p_1 = delta_p(v_1, k=k, d_i=d_i, T_medium=T_average,
                      pressure=pressure, fluid=fluid)

        v_new = 0.5 * (v_1 + v_0)

        p_new = delta_p(v_new, k=k, d_i=d_i, T_medium=T_average,
                        pressure=pressure, fluid=fluid)

        if abs(p_new - p_max) < p_epsilon:
            logger.info("Bi-section method: p_epsilon criterion reached.")
            break

        if abs(v_1 - v_0) < v_epsilon:  # wieso v_1 und v_0?
            logger.info("Bi-section method: v_epsilon criterion reached.")
            break

        else:
            # no stop criteria reached
            # check if p_new is above or below p_max
            if (p_0 - p_max) * (p_new - p_max) < 0:
                v_1 = v_new
            else:
                v_0 = v_new

    logger.info(
        "Maximum flow velocity calculated. Iterations: %d, "
        "Flow velocity: %.4f [m/s], Pressure drop: %.4f [Pa/m]"
        % (n, v_new, p_new)
    )

    return v_new


def calc_power(T_vl=80, T_rl=50, mf=3, p=101325):
    r"""
    Function to calculate the thermal power based on mass flow and temperature difference.

    .. calc_power_equation:

    :math:`P_{th} = \dot{m} \cdot (c_{p_{VL}} \cdot T_{VL} - c_{p_{RL}} \cdot T_{RL})`

    Parameters
    ----------
    T_vl: numeric
        :math:`T_{VL}`: forward temperature [°C]

    T_rl: numeric
        :math:`T_{RL}`: return temperature [C°]

    mf: numeric
        :math:`\dot{m}`: mass flow [kg/s]

    p: numeric
        :math:`p`: pressure [Pa]

    Returns
    -------
    thermal power [W] : numeric

    """
    cp_vl = PropsSI('C', 'T', T_vl + 273.15, 'P', p, 'IF97::Water')

    cp_rl = PropsSI('C', 'T', T_rl + 273.15, 'P', p, 'IF97::Water')

    return mf * (cp_vl * (T_vl + 273.15) - cp_rl * (T_rl + 273.15))


def calc_mass_flow(v, di, T_av, p=101325):
    r"""
    Calculates the mass flow in a pipe for a given density, diameter and flow velocity.
    The average temperature is needed for a correct value of the density.

    .. calc_mass_flow_equation:

    :math:`\dot{m} = \pi \rho_{T_av}  v  \big( \frac{d_i}{2}  \big) ^2`

    Parameters
    ----------
    v : numeric
        :math:`v`: flow velocity [m/s]

    di : numeric
        :math:`d_i`: inner diameter [m]

    T_av : numeric
        :math:`T_av`: temperature level [°C]

    p: numeric
        :math:`p`: pressure [Pa]

    Returns
    -------
    mass flow [kg/s] : numeric

    """
    rho = PropsSI('D', 'T', T_av + 273.15, 'P', p, 'IF97::Water')

    return rho * v * (0.5 * di) ** 2 * math.pi


def calc_mass_flow_P(P, T_av, delta_T, p=101325):
    r"""
    Calculates the mass flow in a pipe for a given power and heat difference.
    The average temperature is needed for a correct value of the heat capacity.

    .. calc_mass_flow_P_equation:

    :math:`\dot{m} = \frac{P}{c_{P_{T_{av}}} \cdot \Delta T}`

    Parameters
    ----------
    P : numeric
        :math:`P`: power [W]

    T_av : numeric
        :math:`T_{av}`: average temperature [°C]

    delta_T : numeric
        :math:`\Delta T`: temperature difference [K]

    p: numeric
        :math:`p`: pressure [Pa]

    Returns
    -------
    mass flow [kg/s]: numeric

    """
    cp = PropsSI('C', 'T', T_av + 273.15, 'P', p, 'IF97::Water')

    return P / (cp * delta_T)


def calc_v_mf(mf, di, T_av, p=101325):
    r"""
    Calculates the flow velocity for a given mass flow and inner diameter.
    The average temperature is needed for a correct value of the density.

    .. calc_v_mf_equation:

    :math:`v = \frac{\dot{m}}{\pi \rho \cdot \big( \frac{d_i}{2} \big)^2 }`

    Parameters
    ----------
    mf : numeric
        :math:`dot{m}`: mass flow [kg/s]

    di : numeric
        :math:`d_i`: inner diameter [m]

    T_av : numeric
        :math:`T_{av}`: average temperature [°C]

     p: numeric
        :math:`p`: pressure [Pa]

    Returns
    -------
    flow velocity [m/s]: numeric

    """
    rho = PropsSI(
        'D', 'T', T_av + 273.15, 'P', p, 'IF97::Water')  # [kg/m^3]

    return mf / (rho * (0.5 * di) ** 2 * math.pi)


def calc_pipe_loss(temp_average, u_value, temp_ground=10):
    r"""
    Calculates the heat loss of a DHS pipe trench.

    Temperatures must be given in the same unit, K or °C.

    .. calc_pipe_loss_equation:

    :math:`P_{loss} = (T_{average} - T_{ground}) \cdot U`

    Parameters
    ----------
    temp_average : float
        :math:`T_{average}`: Average temperature of medium in
        (if u_value relates to forward and return pipe,
        the average temperature of forward and return must be given.)
    u_value : float
        :math:`U`: Heat transmission coefficient of whole trench in W/(m*K)
        (u_value of forward and return pipe must be summed up, if total
        heat loss should be calculated.)
    temp_ground : float
        :math:`T_{ground}`: Temperature of surrounding, e.g. ground.

    Returns
    -------
    Heat loss of pipe trench [W/m]: float
    """
    return (temp_average - temp_ground) * u_value



# -*- coding: utf-8 -*-
"""
These tests test if proper errors are raised when the data is not consistent, of the
wrong type or not all required data are given.

This file is part of project oemof (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/tools/helpers.py

SPDX-License-Identifier: MIT
"""

import copy
import os

import pytest

import dhnx

basedir = os.path.dirname(__file__)

dir_import = os.path.join(basedir, '_files/looped_network_import')

dir_import_inconsistent = os.path.join(basedir, '_files/inconsistent_network_import')

thermal_network = dhnx.network.ThermalNetwork(dir_import)

dir_import_invest = os.path.join(basedir, '_files/investment/')

tn_invest = dhnx.network.ThermalNetwork(dir_import_invest + 'network')

invest_opt = dhnx.input_output.load_invest_options(
    dir_import_invest + 'invest_options'
)


# TODO: The assertions that let these tests fail have yet to be implemented.
#
# def test_datatype_param_nodes():
#     with pytest.raises(TypeError):
#         thermal_network.producers['id'] = np.float(thermal_network.producers['node_id'])
#
#
# def test_datatype_param_pipes():
#     with pytest.raises(TypeError):
#         thermal_network.pipes['id'] = np.float(thermal_network.pipes['id'])
#
#
# def test_required_param_nodes():
#     with pytest.raises(ValueError):
#         thermal_network.producers = thermal_network.producers.drop('lat', axis=1)
#
#
# def test_required_param_pipes():
#     with pytest.raises(ValueError):
#         thermal_network.pipes = thermal_network.pipes.drop('from_node', axis=1)


def test_load_inconsistent_thermal_network():
    with pytest.raises(ValueError):
        dhnx.network.ThermalNetwork(dir_import_inconsistent)


def test_add():
    # missing required attributes
    with pytest.raises(ValueError):
        thermal_network.add('Pipe', 10)


def test_prod_prod():
    # there is a direct producer to producer connection
    with pytest.raises(ValueError, match=r"goes from producers to producers."):
        tn_invest_wrong_1 = copy.deepcopy(tn_invest)
        tn_invest_wrong_1.components['pipes'].at[0, 'to_node'] = 'producers-0'
        dhnx.optimization.optimization_models.setup_optimise_investment(
            tn_invest_wrong_1, invest_opt
        )


def test_cons_cons():
    # there is a edge from consumer to consumer
    with pytest.raises(ValueError, match=r"goes from consumer to consumer"):
        tn_invest_wrong_2 = copy.deepcopy(tn_invest)
        tn_invest_wrong_2.components['pipes'].at[10, 'from_node'] = 'consumers-0'
        dhnx.optimization.optimization_models.setup_optimise_investment(
            tn_invest_wrong_2, invest_opt
        )


def test_prod_cons():
    # there is a direct producer to consumer connection
    with pytest.raises(ValueError, match=r"goes from producers directly to consumers, or vice "):
        tn_invest_wrong_3 = copy.deepcopy(tn_invest)
        tn_invest_wrong_3.components['pipes'].at[0, 'to_node'] = 'consumers-0'
        dhnx.optimization.optimization_models.setup_optimise_investment(
            tn_invest_wrong_3, invest_opt
        )



import matplotlib.pyplot as plt
import dhnx

# Initialize thermal network
network = dhnx.network.ThermalNetwork()
network = network.from_csv_folder('twn_data')

# Load investment parameter
invest_opt = dhnx.input_output.load_invest_options('invest_data')

# plot network
static_map = dhnx.plotting.StaticMap(network)
static_map.draw(background_map=False)
plt.title('Given network')
plt.scatter(network.components.consumers['lon'], network.components.consumers['lat'],
            color='tab:green', label='consumers', zorder=2.5, s=50)
plt.scatter(network.components.producers['lon'], network.components.producers['lat'],
            color='tab:red', label='producers', zorder=2.5, s=50)
plt.scatter(network.components.forks['lon'], network.components.forks['lat'],
            color='tab:grey', label='forks', zorder=2.5, s=50)
plt.text(-2, 32, 'P0', fontsize=14)
plt.text(82, 0, 'P1', fontsize=14)
plt.legend()
plt.show()

# Execute investment optimization
network.optimize_investment(invest_options=invest_opt)

# ####### Postprocessing and Plotting ###########

# get results
results_edges = network.results.optimization['components']['pipes']
print(results_edges[['from_node', 'to_node', 'hp_type', 'capacity',
                     'direction', 'costs', 'losses']])

# print(results_edges[['invest_costs[€]']].sum())
print('Objective value: ', network.results.optimization['oemof_meta']['objective'])

# assign new ThermalNetwork with invested pipes
twn_results = network
twn_results.components['pipes'] = results_edges[results_edges['capacity'] > 0.001]

# plot invested edges
static_map_2 = dhnx.plotting.StaticMap(twn_results)
static_map_2.draw(background_map=False)
plt.title('Result network')
plt.scatter(network.components.consumers['lon'], network.components.consumers['lat'],
            color='tab:green', label='consumers', zorder=2.5, s=50)
plt.scatter(network.components.producers['lon'], network.components.producers['lat'],
            color='tab:red', label='producers', zorder=2.5, s=50)
plt.scatter(network.components.forks['lon'], network.components.forks['lat'],
            color='tab:grey', label='forks', zorder=2.5, s=50)
plt.text(-2, 32, 'P0', fontsize=14)
plt.text(82, 0, 'P1', fontsize=14)
plt.legend()
plt.show()



# -*- coding: utf-8 -*-

"""
This module holds functions for processing the geometry for setting up
the geometry of a ThermalNetwork based on a street geometry and a table of
buildings.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location: https://github.com/oemof/DHNx

This module is not fully tested yet, so use it with care.

SPDX-License-Identifier: MIT
"""
try:
    import geopandas as gpd

except ImportError:
    print("Need to install geopandas to process geometry data.")

try:
    from shapely.geometry import LineString
    from shapely.geometry import MultiPoint
    from shapely.geometry import Point
    from shapely.ops import nearest_points
    from shapely.ops import unary_union
except ImportError:
    print("Need to install shapely to process geometry.")

import logging

import numpy as np
import pandas as pd

from . import geometry_operations as go

logger = logging.getLogger(__name__)  # Create a logger for this module


def line_of_point(point, gdf_lines):
    """Gets index of geometry of a GeoDataFrame, a point is located next to,
      with a distance lower than 1e-8.

    Parameters
    ----------
    point : shapely.geometry.Point
    gdf_lines : geopandas.GeoDataFrame

    Returns
    -------
    int, float or str : Index of GeoDataFrame or Warning, if no geometry found.
    """
    ind = None

    for k, l in gdf_lines.iterrows():

        if l['geometry'].distance(point) < 1e-8:
            ind = k

    if ind is None:
        return Warning('No line found which has point on it!')
    return ind


def point_to_array(point):
    """Returns the coordinates of a point as numpy.array

    Parameters
    ----------
    point : shapely.geometry.Point

    Returns
    -------
    numpy.array()
    """

    return np.array([point.x, point.y])


def calc_lot_foot(line, point):
    """
    Calculates the lot foot point.

    Parameters
    ----------
    line : shapely.geometry.LineString
    point : shapely.geometry.Point

    Returns
    -------
    shapely.geometry.Point
    """
    s_1 = Point(line.coords[0])
    s_2 = Point(line.coords[-1])

    g_1 = point_to_array(s_1)  # end point 1 of line
    g_2 = point_to_array(s_2)  # end point 2 of line

    x_1 = point_to_array(point)

    # calculate lotfusspunkt
    u = g_2 - g_1  # vector of direction
    n = np.array([u[1], -u[0]])  # normal vector of line
    x_0 = g_1  # point on line

    y = x_1 - (np.dot((x_1 - x_0), n) / np.dot(n, n)) * n
    lot_foot_point = Point(y[0], y[1])

    # # alternative generation via intersection
    # # (=> intersections point is not exaclty on lines as well)
    # y = x_1 - 2*(np.dot((x_1 - x_0), n)/np.dot(n, n)) * n
    # lot_line = LineString([(y[0], y[1]), (x_1[0], x_1[1])])
    # lot_foot_point = lot_line.intersection(line)

    return lot_foot_point


def create_object_connections(points, lines, tol_distance=1, n_conn=1,
                              drop_neighbours=True):
    """Connect points to a line network.

    Generally, the nearest point of the next line is used as connection the point.
    Depending on the geometry, there are 3 options, the connection is created:

    - nearest point is line ending => the connection line starts from this line ending

    - nearest point is on the next line:

      a) line endings are outside the tolerance => line is split and the nearest point
      is used as connection point

      b) line endings are within the tolerance distance => the next line ending is
      used as connection point

    The tolerance distance avoids the generation of short line elements.
    This is for example the case if two buildings are directly opposite of the street.
    Using simply the nearest point method could result in very short lines.


    Parameters
    ----------
    points : geopandas.GeoDataFrame
        Points which should be connected to the line. GeoDataFrame with Points as geometry.
    lines : geopandas.GeoDataFrame
        The line-network to which the Points should be connected. The line geometry needs to
        consists of simple lines based on one starting and one ending point. LineStrings
        which contain more than 2 points are not allowed.
    tol_distance : float
        Tolerance distance for choosing the end of the line instead of the nearest point.
    n_conn : int, optional
        Number of connection lines created from each consumer/producer to
        the nearest line segments in the street network. This allows the
        placement of the connection lines to be part of the optimization
        process. The default is 1.
    drop_neighbours : bool, optional
        (Only relevant for n_conn>1). When searching the next connection
        line, ignore the neighbour segements of the last connection segment.
        Chances are that they do not provide an advantage
        over the nearest segment. This allows the next connection
        line to find a more relevant alternative. Default is True.

    Returns
    -------
    geopandas.GeoDataFrame : The newly created connection lines
    geopandas.GeoDataFrame : The updated lines (some lines are split.
        All lines should only touch at the line endings.

    """
    logger.debug("Create connections from street to buildings")

    def create_object_connection(point_geom, lines, tol_distance=tol_distance):
        # Find the nearest line and its nearest point
        lines_merged = unary_union(lines.geometry)
        nearest_line_point = nearest_points(point_geom, lines_merged)[1]
        nearest_line_idx = lines[nearest_line_point.distance(lines.geometry)
                                 < 1e-8].index
        nearest_line = lines.geometry[nearest_line_idx[0]]

        # Check if the nearest point is an end point of the line
        line_start, line_end = nearest_line.boundary.geoms
        if (nearest_line_point.equals(line_start)
           or nearest_line_point.equals(line_end)):
            connection_point = nearest_line_point
        else:
            # Check if the distance of nearest_point_on_line is close
            # to an existing point on the line
            points_on_line = nearest_line.boundary
            closest_existing_point = nearest_points(
                nearest_line_point, points_on_line)[1]
            dist_on_line = nearest_line_point.distance(closest_existing_point)
            if dist_on_line <= tol_distance:
                connection_point = closest_existing_point
                # If connection_point changes, the nearest lines have to be
                # updated. There are probably two nearest lines instead of one
                nearest_line_idx = lines[
                    connection_point.distance(lines.geometry) < 1e-8].index
            else:
                # Split the line and use the nearest point as connection point
                connection_point = nearest_line_point

        # Create connection line (Direction: From street to building)
        conn_line = LineString([connection_point, point_geom])
        return conn_line, nearest_line_idx

    conn_lines_list = []
    for id_full, point_geom in zip(points['id_full'], points.geometry):
        # For each point (building), find the n closest connection lines
        # to the street lines, by dropping the previous closest street
        # sections before searching the next connection line
        lines_drop = []
        conn_lines_id = []
        for i in range(n_conn):
            conn_line, nearest_line_idx = create_object_connection(
                point_geom,
                lines.drop(lines_drop),
                tol_distance=tol_distance)

            lines_drop.extend(nearest_line_idx)
            if drop_neighbours and i + 1 < n_conn:
                # Find all neighbours of the nearest segment(s) and delete them
                # as well. Chances are that they do not provide an advantage
                # over the nearest segment. This allows the next connection
                # line to find a more relevant alternative
                neighbours = lines[lines.touches(
                    unary_union(lines.geometry[nearest_line_idx]))]
                lines_drop.extend(neighbours.index)
                neighbours2 = lines[lines.touches(
                    unary_union(neighbours.geometry))]
                lines_drop.extend(neighbours2.index)

            conn_lines_id.append(conn_line)

        # Create a GeoDataFrame with the connection lines of the current id
        gdf_conn_lines_id = gpd.GeoDataFrame(
            data={'id_full': [id_full] * len(conn_lines_id)},
            geometry=conn_lines_id, crs=lines.crs)

        # Drop duplicate geometries in connection lines
        gdf_conn_lines_id = gdf_conn_lines_id.drop_duplicates(
            subset=gdf_conn_lines_id.geometry.name).reset_index(drop=True)
        # Make sure that the street network is split into new sections
        # where the connection lines meet the street lines
        for conn_line in gdf_conn_lines_id.geometry:
            conn_point = conn_line.boundary.geoms[0]

            if conn_point.within(lines.geometry.boundary).any():
                # Connection point is one of the existing street line points
                continue
            else:
                nearest_line_idx = lines[conn_point.distance(lines.geometry)
                                         < 1e-8].index[0]
                nearest_line = lines.geometry[nearest_line_idx]
                # Identify start and end points for new lines
                line_start, line_end = nearest_line.boundary.geoms
                # Store any existing attributes from original data
                attributes = lines.loc[[nearest_line_idx]].drop(
                    columns=[lines.geometry.name])
                # Remove the line that is to be replaced
                lines.drop([nearest_line_idx], inplace=True)
                # Combine remaining lines with two new replacement lines
                lines = pd.concat([
                    gpd.GeoDataFrame(lines, crs=lines.crs),
                    gpd.GeoDataFrame(
                        geometry=[LineString([line_start, conn_point]),
                                  LineString([conn_point, line_end])],
                        crs=lines.crs,
                        data=pd.concat([attributes, attributes]))],
                    ignore_index=True)

        conn_lines_list.append(gdf_conn_lines_id)

    gdf_conn_lines = pd.concat(conn_lines_list, ignore_index=True)

    return gdf_conn_lines, lines


def check_geometry_type(gdf, types):
    """
    Checks, if a geodataframe has only the given geometry types in its GeoSeries.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        DataFrame to be checked.
    types : list
        List of types allowed for GeoDataFrame.

    """
    actual_types = set(gdf['geometry'].type)

    for type in actual_types:
        if type not in types:
            raise TypeError(
                "Your input geometry has the wrong type. "
                "Expected: {}. Got: {}".format(types, type)
            )


def create_points_from_polygons(gdf, method='midpoint'):
    """
    Converts the geometry of a polygon layer to a point layer.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
    method : str
        Method to create a point from a polygon.

    Returns
    -------
    geopandas.GeoDataFrame : GeoDataFrame with a point geometry.

    """

    if gdf['geometry'].values[0].geom_type == 'Point':
        return gdf

    if method == 'midpoint' or method == 'boundary':
        # (method 'boundary' is performed later and needs the centroid)
        gdf['geometry'] = gdf['geometry'].centroid
        return gdf

    raise ValueError(
        "No other methods than 'midpoint' and 'boundary' implemented!"
    )


def run_point_method_boundary(consumers_poly, consumers, lines_consumers):
    """Run 'boundary' method for finding the building connection point.

    This is meant to be called once with the consumers as input and
    once with the producers (producers_poly, producers, lines_producers).

    The 'midpoint' method (using the centroid) must already have been run,
    generating the default connection lines from street to centroid.

    If there is only one intersection between that line and the boundary
    of the building, this intersection point is used as the connection
    point instead (and the connection line is shortened accordingly).

    However, complex building shapes can produce multiple intersections. In
    this case, the intersection with the 'convex hull' of the building is used
    instead. This may result in connection points that do not touch an
    actual building wall, but it should still be an improvement compared to
    the 'midpoint' method.

    In case of no intersections with the building boundary (possible for e.g.
    U-shaped buildings), the original centroid is used. This is also
    necessary when the the street line already touches the building, to
    prevent deleting the connection line.

    Parameters
    ----------
    consumers_poly : geopandas.GeoDataFrame
        Polygons of the consumer buildings. Point geometries are also allowed,
        but they are not changed.
    consumers : geopandas.GeoDataFrame
        Points of the consumer buildings (as returned by 'midpoint' method).
    lines_consumers : geopandas.GeoDataFrame
        Connection lines from street to each consumer point.

    Returns
    -------
    consumers : geopandas.GeoDataFrame
        Updated points of the consumer buildings.
    lines_consumers : geopandas.GeoDataFrame
        Updated connection lines from street to each consumer point.

    """
    logger.info('Run "boundary" method for finding the building connections')
    # lines_consumers may represent multiple lines per consumer
    # Duplicate geometries in consumers_poly have to be created accordingly
    consumers_poly['id_full'] = consumers['id_full']
    consumers_poly = pd.merge(
        left=consumers_poly,
        right=lines_consumers.drop(columns=[lines_consumers.geometry.name]),
        how='right', on='id_full')

    # When using the original consumer points as a fallback later, we
    # require it to have the same index as lines_consumers. Therefore
    # we create the 'duplicate' consumers object
    consumers_d = pd.merge(
        left=consumers,
        right=lines_consumers.drop(columns=[lines_consumers.geometry.name]),
        how='right', on='id_full')

    # Cut the part off of each "line_consumer" that is within the building
    # polygon. As a result, the heating grid will only reach to the wall of
    # the building.
    lines_consumers_n = gpd.GeoDataFrame(
        geometry=lines_consumers.difference(consumers_poly, align=False))
    # This produces a MultiLineString for complex building polygons, where
    # the boundary and the simple lines from centroid to street intersect
    # multiple times. This would not be a valid connection line. In those
    # cases the 'convex hull' of the building is used instead.
    lines_consumers_n.loc[lines_consumers_n.type == "MultiLineString"] = \
        gpd.GeoDataFrame(geometry=lines_consumers.difference(
            consumers_poly.convex_hull, align=False))
    # Only keep the new consumer lines if they have a useful minimum length.
    # There was an edgecase where a street 'almost' touched a building,
    # and the cut consumer line had a length of 1e-9 m
    lines_consumers_n[lines_consumers_n.length < 1e-3] = LineString()

    # Now the "consumers" (point objects for each building) need to be
    # updated to touch the end of the consumer_lines
    consumers_n = gpd.GeoDataFrame(geometry=lines_consumers.intersection(
        consumers_poly.boundary, align=False))
    consumers_n.loc[consumers_n.type == "MultiPoint"] = \
        gpd.GeoDataFrame(geometry=lines_consumers.intersection(
            consumers_poly.convex_hull.boundary, align=False))

    # Sometimes the centroid does not lie within a building and there may be
    # no intersetions, i.e. the new point is an 'empty' geometry. This can
    # happen if buildings are multipolygons, which is not forbidden.
    # Sometimes the new lines are empty (e.g. because a street and a building
    # object cross each other).
    # In these cases the original geometry is used for points and lines.
    mask1 = (consumers_n.is_empty | lines_consumers_n.is_empty)

    # Another special case has to be covered. If the original street lines
    # already touch the building wall, no additional connection line would be
    # necessary. However, in dhnx each building needs one connection line.
    # Thus the geometry from the 'midpoint' method is used here, too.
    # Find the problematic cases by testing if the new connection point
    # equals the starting point of the connection line.
    mask2 = consumers_n.geom_equals(
        lines_consumers.geometry.apply(lambda line: line.boundary.geoms[0]))

    # If for whatever reason the street-side "start" of the new connection
    # line is not the same point as the original connection line start, use
    # the original line. This may happen for complex geometries, where the
    # street line lies within the building geometry
    lines_consumers_n_start = lines_consumers_n.copy()
    lines_consumers_n_start.geometry = (
        lines_consumers_n_start[~mask1 & ~mask2].boundary.apply(
            lambda g: g.geoms[0]))
    mask3 = lines_consumers_n_start.geom_equals(
        lines_consumers.geometry.apply(lambda line: line.boundary.geoms[0]))

    # Now apply all the filters above to reset the geometries
    mask = mask1 | mask2 | ~mask3
    consumers_n.loc[mask] = consumers_d.loc[mask].geometry
    lines_consumers_n.loc[mask] = lines_consumers.loc[mask].geometry

    # Now update all the original variables with the new data
    lines_consumers_n['id_full'] = lines_consumers['id_full']
    consumers_n['id_full'] = lines_consumers_n['id_full']

    # If multiple building connection lines existed before, we now also
    # have created multiple building points for each building.
    # We need to group those points into one MultiPoint per initial unique
    # building, to keep the original index structure intact.
    consumers_n = (pd.DataFrame(consumers_n)  # convert gdf to df
                   .groupby('id_full', sort=False, as_index=False)
                   .agg(lambda x: MultiPoint(x.values))  # returns df
                   .set_geometry(consumers_n.geometry.name,  # convert to gdf
                                 crs=consumers_n.crs)
                   )

    # For each new consumer point(s), test if they actually touch
    # the new conumser line(s) that have the same 'id_full' assigned
    for id_full, points in zip(consumers_n['id_full'], consumers_n.geometry):
        if not points.touches(lines_consumers_n[
                lines_consumers_n['id_full'] == id_full].geometry).all():
            raise ValueError(f"Points from {id_full} have no matching lines")

    consumers.geometry = consumers_n.geometry
    lines_consumers = lines_consumers_n
    return consumers, lines_consumers


def check_duplicate_geometries(gdf):
    """Test the input GeoDataFrame for duplicate geometries and plot them."""
    if gdf.duplicated(subset="geometry").any():
        idx = gdf.duplicated(subset="geometry")
        try:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(dpi=400)
            gdf.loc[~idx].plot(ax=ax, color='green')
            gdf.loc[idx].plot(ax=ax, color='red')
            plt.title("Red are duplicate geometries. Please fix!")
            plt.show()
        except ImportError:
            logger.info("Install matplotlib to show a plot of the duplicate "
                        "geometries.")
        raise ValueError("GeoDataFrame has {} duplicate geometries"
                         .format(len(gdf.loc[idx])))


def process_geometry(lines, consumers, producers,
                     method='midpoint', projected_crs=4647,
                     tol_distance=2, reset_index=True, n_conn=1, n_conn_prod=1,
                     welding=True):
    """
    This function connects the consumers and producers to the line network, and prepares the
    attributes of the geopandas.GeoDataFrames for importing as dhnx.ThermalNetwork.

    The ids of the lines are overwritten.

    Parameters
    ----------
    lines : geopandas.GeoDataFrame
        Potential routes for the DHS. Expected geometry Linestrings or MultilineStrings.
        The graph of this line network should be connected.
    consumers : geopandas.GeoDataFrame
        Location of demand/consumers. Expected geometry: Polygons or Points.
    producers : geopandas.GeoDataFrame
        Location of supply sites. Expected geometry: Polygons or Points.
    method : str
        Method for creating the point if polygons are given for the consumers
        and producers. Method 'midpoint' uses the centroid of each building
        polygon. Method 'boundary' moves the point to the boundary (wall) of
        the building, along the line constructed from centroid to the street.
    multi_connections : bool
        Setting if a building should be connected to multiple streets.
    projected_crs : EPSG integer code
        EPSG Coordinate reference system number (eg 4647),
        which is used for the geometry operations.
        A projected crs must be used!
    tol_distance : float
        Tolerance distance at connection the points to the line network
        for choosing the end of the line instead of the lot.
    reset_index : bool
        If True, reset the index and ignore the existing index. If False,
        use the existing index for consumer and producer identificators.
        Default: True
    n_conn : int, optional
        Number of connection lines created from each consumer to
        the nearest line segments in the street network. This allows the
        placement of the connection lines to be part of the optimization
        process. The default is 1.
    n_conn_prod : int, optional
        Number of connection lines created from each producer to
        the nearest line segments in the street network. This allows the
        placement of the connection lines to be part of the optimization
        process. The default is 1.
    welding : bool, optional
        Weld continuous line segments together and cut loose ends. This
        can improve the performance of the optimization, as it decreases
        the total number of line elements. Default is True.

    Returns
    -------
    dict : Dictionary with 4 geopandas.GeoDataFrames: The keys of the Dict are
           equal to the components of the dhnx.ThermalNetwork: 'forks', 'consumers',
           'producers', 'pipes'.

    """
    if not reset_index:
        raise ValueError("Keeping the orginal index is currently not "
                         "supported. Use 'reset_index=True'.")

    # Copies of the original polygons are needed for method 'boundary'
    consumers_poly = go.check_crs(consumers, crs=projected_crs).copy()
    producers_poly = go.check_crs(producers, crs=projected_crs).copy()

    # check whether the expected geometry is used for geo dataframes
    check_geometry_type(lines, types=['LineString', 'MultiLineString'])
    for gdf in [producers, consumers, producers_poly, consumers_poly]:
        check_geometry_type(gdf, types=['Polygon', 'Point', 'MultiPolygon'])
        check_duplicate_geometries(gdf)

    # split multilinestrings to single lines with only 1 starting and 1 ending point
    lines = go.split_multilinestr_to_linestr(lines)

    # check and convert crs if it is not already the `projected_crs`
    lines = go.check_crs(lines, crs=projected_crs)

    for layer in [producers, consumers]:
        layer = go.check_crs(layer, crs=projected_crs)
        layer = create_points_from_polygons(layer, method=method)
        layer['lat'] = layer['geometry'].apply(lambda x: x.y)
        layer['lon'] = layer['geometry'].apply(lambda x: x.x)

    for layer in [producers, consumers, producers_poly, consumers_poly]:
        if reset_index:
            layer.reset_index(inplace=True, drop=True)
            layer.index.name = 'id'
            layer.drop(columns=['id'], inplace=True, errors='ignore')
        else:
            if layer.index.has_duplicates:
                raise ValueError("The index of input data has duplicate "
                                 "values, which is not allowed")

    producers['id_full'] = 'producers-' + producers.index.astype('str')
    producers['type'] = 'G'
    consumers['id_full'] = 'consumers-' + consumers.index.astype('str')
    consumers['type'] = 'H'

    # Add lines to consumers and producers
    lines_consumers, lines = create_object_connections(
        consumers, lines, tol_distance=tol_distance, n_conn=n_conn)
    lines_producers, lines = create_object_connections(
        producers, lines, tol_distance=tol_distance, n_conn=n_conn_prod)
    if not reset_index:
        # Connection lines are ordered the same as points. Match their indexes
        lines_consumers.index = consumers.index
        lines_producers.index = producers.index

    if method == 'boundary':
        # Can only be performed after 'midpoint' method
        consumers, lines_consumers = run_point_method_boundary(
            consumers_poly, consumers, lines_consumers)
        producers, lines_producers = run_point_method_boundary(
            producers_poly, producers, lines_producers)

    if welding:
        # Weld continuous line segments together and cut loose ends
        lines = go.weld_segments(
            lines, lines_producers, lines_consumers,
            # debug_plotting=True,
        )

    # Keep only the shortest of all lines connecting the same two points
    lines = go.drop_parallel_lines(lines)

    # add additional line identifier
    lines_producers['type'] = 'GL'  # GL for generation line
    lines['type'] = 'DL'  # DL for distribution line
    lines_consumers['type'] = 'HL'  # HL for house line

    # generate forks point layer
    forks = go.create_forks(lines)

    # concat lines
    lines_all = pd.concat([lines, lines_consumers, lines_producers], sort=False)
    lines_all.reset_index(inplace=True, drop=True)
    if reset_index:
        lines_all.index.name = 'id'
        lines_all.drop(columns=['id'], inplace=True, errors='ignore')

    # concat point layer
    points_all = pd.concat([
        consumers[['id_full', 'geometry']],
        producers[['id_full', 'geometry']],
        forks[['id_full', 'geometry']]],
        sort=False
    )

    # add from_node, to_node to lines layer
    lines_all = go.insert_node_ids(lines_all, points_all)

    lines_all['length'] = lines_all.length
    logger.info(
        "Total line length is {:.0f} m".format(lines_all['length'].sum()))

    # Convert all MultiLineStrings to LineStrings
    check_geometry_type(lines_all, types=['LineString'])

    # ## check for near points
    go.check_double_points(points_all, id_column='id_full')

    return {
        'forks': forks,
        'consumers': consumers,
        'producers': producers,
        'pipes': lines_all,
    }



import pytest

from dhnx.optimization.precalc_hydraulic import calc_d_p
from dhnx.optimization.precalc_hydraulic import calc_k_v
from dhnx.optimization.precalc_hydraulic import calc_lambda_laminar
from dhnx.optimization.precalc_hydraulic import calc_lambda_rough
from dhnx.optimization.precalc_hydraulic import calc_lambda_transition
from dhnx.optimization.precalc_hydraulic import calc_lambda_turb1
from dhnx.optimization.precalc_hydraulic import calc_lambda_turb2
from dhnx.optimization.precalc_hydraulic import calc_lambda_turb3
from dhnx.optimization.precalc_hydraulic import calc_mass_flow
from dhnx.optimization.precalc_hydraulic import calc_mass_flow_P
from dhnx.optimization.precalc_hydraulic import calc_pipe_loss
from dhnx.optimization.precalc_hydraulic import calc_power
from dhnx.optimization.precalc_hydraulic import calc_Re
from dhnx.optimization.precalc_hydraulic import calc_v
from dhnx.optimization.precalc_hydraulic import calc_v_mf
from dhnx.optimization.precalc_hydraulic import delta_p
from dhnx.optimization.precalc_hydraulic import v_max_bisection
from dhnx.optimization.precalc_hydraulic import v_max_secant


def test_calc_v():
    v = calc_v(720, 0.5)
    assert round(v, 5) == 1.01859


def test_calc_power():
    P_th = calc_power(80, 50, 3)
    assert round(P_th, 5) == 393070.3261


def test_calc_mass_flow():
    mf = calc_mass_flow(3, 0.5, 20)
    assert round(mf, 5) == 587.99192


def test_calc_mass_flow_P():
    mf = calc_mass_flow_P(50000, 20, 5)
    assert round(mf, 5) == 2.3896


def test_calc_v_mf():
    v = calc_v_mf(200, 0.5, 20)
    assert round(v, 5) == 1.02042


def test_calc_pipe_loss():
    loss = calc_pipe_loss(20, 0.1, 10)
    assert round(loss, 5) == 1


def test_calc_k_v():
    k_v = calc_k_v(5, 4)
    assert k_v == 1.25


def test_calc_Re():
    Re = calc_Re(1, 2, 4)
    assert Re == 0.5


def test_calc_lam_laminar():
    lam = calc_lambda_laminar(64)
    assert lam == 1


def test_calc_d_p():
    dp = calc_d_p(1, 1, 2, 1, 2)
    assert dp == 1


def test_calc_lam_turb1():
    lam = calc_lambda_turb1(0.001)
    assert round(lam, 5) == 1.77925


def test_calc_lam_turb2():
    lam = calc_lambda_turb2(1)
    assert lam == 0.2242


def test_calc_lam_rough():
    lam = calc_lambda_rough(1, 1)
    assert round(lam, 5) == 0.77116


def test_calc_lam_turb3():
    lam = calc_lambda_turb3(1)
    assert round(lam, 5) == 12.18494


def test_calc_lam_transition():
    lam = calc_lambda_transition(1, 1, 1)
    assert round(lam, 5) == 23.41445


def test_wrong_initial_guess_error():
    with pytest.raises(AttributeError, match=r"initial guesses `v_0` and "):
        v_max_bisection(0.1, 65, v_0=1, v_1=1)


def test_bisection_method_velocity():
    bi_1 = v_max_bisection(
        0.1, 65, k=0.1, p_max=100,
        p_epsilon=0.1, v_epsilon=0.001, v_0=1, v_1=0.1,
        pressure=101325, fluid='IF97::Water'
    )
    assert round(bi_1, 7) == 0.9876953


def test_secant_method_velocity():
    se_1 = v_max_secant(
        0.4, 80, k=0.1, p_max=250, p_epsilon=1,
        v_0=1, v_1=2,
        pressure=101325, fluid='IF97::Water'
    )
    assert round(se_1, 7) == 3.7593294


def test_delta_p1():  # laminar
    dp = delta_p(1E-6, 1)
    assert round(dp, 13) == 1.00538E-8


def test_delta_p2():  # turb, Re < 10**5
    dp = delta_p(1, 5E-3, k=0.01)
    assert round(dp, 5) == 2743.41722


def test_delta_p3():  # turb, 10**5 < Re < 10**6
    dp = delta_p(10, 5E-3, k=0.001)
    assert round(dp, 5) == 156745.5396


def test_delta_p4():  # turb, Re > 10**6
    dp = delta_p(100, 5E-3, k=0.0001)
    assert round(dp, 5) == 10456325.31958


def test_delta_p5():  # turb, Re*k/di > 1300
    dp = delta_p(100, 5E-3, k=0.01)
    assert round(dp, 5) == 22592027.65789


def test_delta_p6():  # turb, transition
    dp = delta_p(100, 5E-3, k=0.0003)
    assert round(dp, 5) == 11865210.59373



# -*- coding: utf-8 -*-
"""
These tests test if proper errors are raised when the data is not consistent, of the
wrong type or not all required data are given.

This file is part of project oemof (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/tools/helpers.py

SPDX-License-Identifier: MIT
"""

import geopandas as gpd
from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import Point

from dhnx.gistools import connect_points as cp
from dhnx.gistools import geometry_operations as go


def test_lot_foot_calc():
    point = Point([(0.5, 1)])
    line = LineString([(0, 0), (1, 1)])
    assert cp.calc_lot_foot(line, point) == Point([(0.75, 0.75)])


def test_split_linestring():
    line1 = LineString([(0, 0), (1, 3), (2, 0)])
    line2 = MultiLineString(lines=[line1, LineString([(5, 5), (7, 9), (3, 4)])])
    line3 = LineString([(1, 1), (5, 1)])
    gdf_line = gpd.GeoDataFrame(geometry=[line1, line2, line3])
    results = go.split_multilinestr_to_linestr(gdf_line)
    assert gdf_line.geometry.length.sum() == results.length.sum()
    assert len(results.index) == 7



from dhnx.network import ThermalNetwork

# Initialize a thermal networks
thermal_network = ThermalNetwork()

# Load data from csv
input_data_dir = 'tree'

thermal_network.from_csv_folder(input_data_dir)

# Create simulation model and save results
thermal_network.simulate(results_dir='tree_results')

# Print results
print('================================================================')
for k, v in thermal_network.results['simulation'].items():
    print(k)
    print('----------------------------------------------------------------')
    print(v, '\n')
    print('================================================================')



# -*- coding: utf-8 -*-
"""
These tests test little examples for the use of the library.

This file is part of project oemof (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/tools/helpers.py

SPDX-License-Identifier: MIT
"""

import os

import networkx as nx
import pandas as pd

import dhnx

from . import helpers

tmpdir = helpers.extend_basic_path('tmp')

basedir = os.path.dirname(__file__)

dir_import_tree = os.path.join(basedir, '_files/tree_network_import')

dir_import_looped = os.path.join(basedir, '_files/looped_network_import')

tree_thermal_network = dhnx.network.ThermalNetwork(dir_import_tree)

looped_thermal_network = dhnx.network.ThermalNetwork(dir_import_looped)

dir_import_invest = os.path.join(basedir, '_files/investment/')

tn_invest = dhnx.network.ThermalNetwork(dir_import_invest + 'network')

invest_opt = dhnx.input_output.load_invest_options(
    dir_import_invest + 'invest_options'
)


def test_import_export_csv():
    dir_export = os.path.join(tmpdir, 'network_export')

    network = dhnx.network.ThermalNetwork()
    network = network.from_csv_folder(dir_import_looped)

    network.to_csv_folder(dir_export)

    helpers.check_if_csv_dirs_equal(dir_import_looped, dir_export)


def test_access_attributes():

    network = dhnx.network.ThermalNetwork(dir_import_looped)

    assert isinstance(network.available_components, pd.DataFrame)

    assert network.component_attrs.consumers.id.type == 'int'

    assert isinstance(network.components.consumers, pd.DataFrame)

    assert isinstance(network.sequences.producers.temp_inlet, pd.DataFrame)


def test_get_nx_graph():

    nx_graph = looped_thermal_network.to_nx_graph()

    assert isinstance(nx_graph, nx.Graph)


def test_static_map():
    # plot static map
    dhnx.plotting.StaticMap(looped_thermal_network)


def test_interactive_map():
    # plot interactive map
    interactive_map = dhnx.plotting.InteractiveMap(looped_thermal_network)
    interactive_map.draw()


def test_setup_operation_optimization():

    looped_thermal_network.optimize_operation()


def test_setup_investment_optimization():

    dhnx.optimization_models.setup_optimise_investment(tn_invest, invest_opt)


def test_setup_simulation():

    tree_thermal_network.simulate()



# -*- coding: utf-8

"""
This module is designed to hold optimization model implementations. The
implementation makes use of oemof-solph.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""
# pylint: skip-file
import warnings
from collections import namedtuple

from oemof.network import Node
from oemof.solph import Investment
from oemof.solph._plumbing import sequence
from pyomo.core.base.block import ScalarBlock
from pyomo.environ import Constraint
from pyomo.environ import NonNegativeReals
from pyomo.environ import Set
from pyomo.environ import Var


class Label(namedtuple('solph_label', ['tag1', 'tag2', 'tag3', 'tag4'])):
    __slots__ = ()

    def __str__(self):
        """The string is used within solph as an ID, so it hast to be unique"""
        return '_'.join(map(str, self._asdict().values()))


class HeatPipeline(Node):
    r"""A HeatPipeline represent a Pipeline in a district heating system.

    This is done by a Transformer with a constant energy loss independent of
    actual power, but dependent on the nominal power.
    The HeatPipeline is a single-input-single-output transformer.

    Parameters
    ----------
    length : float
        Length of HeatPipeline.
    heat_loss_factor : float
        Heat loss per length unit as fraction of the nominal power. Can also be
        defined by a series.


    Note
    ----
    This component is experimental. Use it with care. See also
    :py:class:`~oemof.solph.network.Transformer`.


    The following sets, variables, constraints and objective parts are created
     * :py:class:`~dhnx.optimization.oemof_heatpipe.HeatPipelineBlock` (if no
       Investment object present)
     * :py:class:`~dhnx.optimization.oemof_heatpipe.HeatPipelineInvestBlock`
      (ifInvestment object present)

    Examples
    --------
    # TODO : example

    """

    def __init__(
        self,
        inputs,
        outputs,
        label=None,
        heat_loss_factor=0,
        heat_loss_factor_fix=0,
    ):

        self.heat_loss_factor = sequence(heat_loss_factor)
        self.heat_loss_factor_fix = sequence(heat_loss_factor_fix)

        super().__init__(
            inputs=inputs,
            outputs=outputs,
            label=label,
        )

        self._invest_group = False
        self._nonconvex_group = False
        self._demand_group = False

        if len(self.inputs) > 1 or len(self.outputs) > 2:
            if len(self.outputs) == 2:
                self._demand_group = True
            else:
                raise ValueError("Heatpipe must not have more than"
                                 " one input and two outputs!")

        for f in self.inputs.values():
            if f.nonconvex is not None:
                raise ValueError(
                    "Inputflow must not be of type NonConvexFlow!")

        for f in self.outputs.values():
            if f.nonconvex is not None:
                self._nonconvex_group = True

        self._check_flows_invest()

        if (self._nonconvex_group is True) and (self._invest_group is True):
            raise ValueError(
                "Either an investment OR a switchable heatloss can be set"
                " (NonConvexFlow)."
                " Remove the NonConvexFlow or drop "
                "the Investment attribute.")

        if self._invest_group is True:
            self._set_flows()
            o = list(self.outputs.keys())[0]
            if (self.heat_loss_factor_fix[0] > 0) \
                    and (self.outputs[o].investment.nonconvex is False):
                warnings.warn(
                    "Be careful! In case of a convex Investment "
                    "(Investment.nonconvex is False), the "
                    "'heat_loss_factor_fix' is considered, even though the "
                    "investment might be zero! => A simple sink could be "
                    "the results. Hopefully, you know what you are doing.")
        else:
            self._set_nominal_value()

    def _check_flows_invest(self):
        # for flow in self.inputs.values():
        #     if isinstance(flow.investment, Investment):
        #         raise ValueError(
        #             "The investment must be defined at the Outputflow!")

        for flow in self.outputs.values():
            if isinstance(flow.investment, Investment):
                self._invest_group = True

    def _set_flows(self):
        # sets the input flow to investment
        for flow in self.inputs.values():
            flow.investment = Investment()

    # set nominal values of in- and outflow equal in case of
    # invest_group = False
    def _set_nominal_value(self):
        i = list(self.inputs.keys())[0]
        o = list(self.outputs.keys())[0]
        if self.outputs[o].nominal_value is not None:
            self.inputs[i].nominal_value = \
                self.outputs[o].nominal_value
        elif self.inputs[i].nominal_value is not None:
            self.outputs[o].nominal_value = \
                self.inputs[i].nominal_value

    def constraint_group(self):
        if self._invest_group is True:
            return HeatPipelineInvestBlock
        return HeatPipelineBlock


class HeatPipelineBlock(ScalarBlock):  # pylint: disable=too-many-ancestors
    r"""Block representing a pipeline of a district heating system.
    :class:`~dhnx.optimization.oemof_heatpipe.HeatPipeline`

    **The following constraints are created:**

    # TODO : Check and fix equations!

    .. _HeatPipelineBlock-equations:

    .. math:: \dot{Q}_{out}(t) = \dot{Q}_{in}(t) \cdot
        \frac{\eta_{out}}{\eta_{in}} - \dot{Q}_{loss}(t)

    .. math:: \dot{Q}_{loss}(t) = f_{loss}(t) \cdot l \cdot \dot{Q}_{nominal}

    The symbols used are defined as follows
    (with Variables (V) and Parameters (P)):

    .. csv-table::
        :header: "symbol", "attribute", "type", "explanation"
        :widths: 1, 1, 1, 1

        ":math:`\dot{Q}_{out}(t)`", ":py:obj:`flow[n, o, t]`", "V", "Heat
        output"
        ":math:`\dot{Q}_{in}(t)`", ":py:obj:`flow[i, n, t]`", "V", "Heat input"
        ":math:`\dot{Q}_{loss}(t)`", ":py:obj:`heat_loss[n, t]`", "P", "Heat
        loss of heat pipeline"
        ":math:`\dot{Q}_{nominal}`", ":py:obj:`flows[n, o].nominal_value`", "
        P", "Nominal capacity of heating pipeline"
        ":math:`f_{loss}(t)`", ":py:obj:`heat_loss_factor`", "P", "Specific
        heat loss factor for pipeline"
        ":math:`l`", ":py:obj:`length`", "P", "Length of heating pipeline"


    """

    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """ Creates the linear constraint for the class:`Heatpipe`
        block.

        Parameters
        ----------
        group : list

        """
        if group is None:
            return None

        m = self.parent_block()

        self.HEATPIPES = Set(initialize=list(group))
        self.CONVEX_HEATPIPES = Set(
            initialize=[n for n in group
                        if n.outputs[list(n.outputs.keys())[0]].nonconvex
                        is None]
        )
        self.NONCONVEX_HEATPIPES = Set(
            initialize=[n for n in group
                        if n.outputs[list(n.outputs.keys())[0]].nonconvex
                        is not None])

        # Defining Variables
        self.heat_loss = Var(self.HEATPIPES, m.TIMESTEPS,
                             within=NonNegativeReals)

        def _heat_loss_rule_fix(block, n, t):
            """Rule definition for the heat loss depending on the nominal
            capacity for fix fix heat loss.
            """
            o = list(n.outputs.keys())[0]

            expr = 0
            expr += - block.heat_loss[n, t]
            expr +=\
                n.heat_loss_factor[t] * m.flows[n, o].nominal_value
            expr += n.heat_loss_factor_fix[t]
            return expr == 0

        self.heat_loss_equation_fix = Constraint(
            self.CONVEX_HEATPIPES, m.TIMESTEPS, rule=_heat_loss_rule_fix)

        def _heat_loss_rule_on_off(block, n, t):
            """Rule definition for the heat loss depending on the nominal
            capacity. Here, the losses can be "switched off".
            """
            o = list(n.outputs.keys())[0]

            expr = 0
            expr += - block.heat_loss[n, t]
            expr += (
                n.heat_loss_factor[t] * m.flows[n, o].nominal_value
                + n.heat_loss_factor_fix[t]
            ) * m.NonConvexFlowBlock.status[n, o, t]
            return expr == 0

        self.heat_loss_equation_on_off = Constraint(
            self.NONCONVEX_HEATPIPES, m.TIMESTEPS, rule=_heat_loss_rule_on_off)

        def _relation_rule(block, n, t):
            """Link input and output flow and subtract heat loss."""
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]

            expr = 0
            try:  # oemof.solph<=0.5.0
                expr += - m.flow[n, o, t]
                expr += m.flow[i, n, t]
            except KeyError:  # oemof.solph>=0.5.1
                period = 0  # Periods are not (yet) supported in DHNx
                expr += - m.flow[n, o, period, t]
                expr += m.flow[i, n, period, t]

            expr += - block.heat_loss[n, t]
            return expr == 0

        self.relation = Constraint(self.HEATPIPES, m.TIMESTEPS,
                                   rule=_relation_rule)


class HeatPipelineInvestBlock(ScalarBlock):  # pylint: disable=too-many-ancestors
    r"""Block representing a pipeline of a district heating system.
    :class:`~oemof.solph.custom.HeatPipeline`

    **The following constraints are created:**

    # TODO : Check and fix equations!

    .. _HeatPipelineInvestBlock-equations:

    .. math:: \dot{Q}_{out}(t) = \dot{Q}_{in}(t) \cdot
        \frac{\eta_{out}}{\eta_{in}} - \dot{Q}_{loss}(t)

    .. math:: \dot{Q}_{loss}(t) = f_{loss}(t) \cdot l \cdot \dot{Q}_{nominal}

    The symbols used are defined as follows
    (with Variables (V) and Parameters (P)):

    .. csv-table::
        :header: "symbol", "attribute", "type", "explanation"
        :widths: 1, 1, 1, 1

        ":math:`\dot{Q}_{out}(t)`", ":py:obj:`flow[n, o, t]`", "V", "Heat
        output"
        ":math:`\dot{Q}_{in}(t)`", ":py:obj:`flow[i, n, t]`", "V", "Heat input"
        ":math:`\dot{Q}_{loss}(t)`", ":py:obj:`heat_loss[n, t]`", "V", "Heat
        loss of heat pipeline"
        ":math:`\dot{Q}_{nominal}`", ":py:obj:`flows[n, o].nominal_value`", "
        V", "Nominal capacity of heating pipeline"
        ":math:`f_{loss}(t)`", ":py:obj:`heat_loss_factor`", "P", "Specific
        heat loss factor for pipeline"
        ":math:`l`", ":py:obj:`length`", "P", "Length of heating pipeline"

    """

    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """ Creates the linear constraint for the class:`HeatPipeline`
        block.

        Parameters
        ----------
        group : list

        """
        if group is None:
            return None

        m = self.parent_block()

        # Defining Sets
        self.INVESTHEATPIPES = Set(initialize=list(group))
        self.CONVEX_INVESTHEATPIPES = Set(
            initialize=[n for n in group
                        if n.outputs[
                            list(n.outputs.keys())[0]
                        ].investment.nonconvex is False]
        )
        self.NONCONVEX_INVESTHEATPIPES = Set(
            initialize=[n for n in group
                        if n.outputs[
                            list(n.outputs.keys())[0]
                        ].investment.nonconvex is True]
        )

        self.INVESTHEATPIPES_NO_DEMAND = Set(
            initialize=[n for n in group if len(n.outputs.keys()) == 1])
        self.INVESTHEATPIPES_WITH_DEMAND = Set(
            initialize=[n for n in group if len(n.outputs.keys()) == 2])

        # Defining Variables
        self.heat_loss = Var(self.INVESTHEATPIPES, m.TIMESTEPS,
                             within=NonNegativeReals)

        def _heat_loss_rule_convex(block, n, t):
            """Rule definition for constraint to connect the installed capacity
            and the heat loss
            """
            expr = 0
            expr += - block.heat_loss[n, t]
            try:  # oemof.solph<=0.5.0
                expr += n.heat_loss_factor[t] * m.InvestmentFlowBlock.invest[
                    n, list(n.outputs.keys())[0],
                ]
            except KeyError:  # oemof.solph>=0.5.1
                period = 0  # Periods are not (yet) supported in DHNx
                expr += n.heat_loss_factor[t] * m.InvestmentFlowBlock.invest[
                    n, list(n.outputs.keys())[0], period
                ]
            expr += n.heat_loss_factor_fix[t]
            return expr == 0
        self.heat_loss_equation_convex = Constraint(
            self.CONVEX_INVESTHEATPIPES, m.TIMESTEPS,
            rule=_heat_loss_rule_convex
        )

        def _heat_loss_rule_nonconvex(block, n, t):
            """Rule definition for constraint to connect the installed capacity
            and the heat loss
            """
            expr = 0
            expr += - block.heat_loss[n, t]
            try:  # oemof.solph<=0.5.0
                expr += n.heat_loss_factor[t] * m.InvestmentFlowBlock.invest[
                    n, list(n.outputs.keys())[0]]
                expr += n.heat_loss_factor_fix[t] * \
                    m.InvestmentFlowBlock.invest_status[
                        n, list(n.outputs.keys())[0]]
            except KeyError:  # oemof.solph>=0.5.1
                period = 0  # Periods are not (yet) supported in DHNx
                expr += n.heat_loss_factor[t] * m.InvestmentFlowBlock.invest[
                    n, list(n.outputs.keys())[0], period]
                expr += n.heat_loss_factor_fix[t] * \
                    m.InvestmentFlowBlock.invest_status[
                        n, list(n.outputs.keys())[0], period]
            return expr == 0

        self.heat_loss_equation_nonconvex = Constraint(
            self.NONCONVEX_INVESTHEATPIPES, m.TIMESTEPS,
            rule=_heat_loss_rule_nonconvex
        )

        def _relation_rule_no_demand(block, n, t):
            """Link input and output flow and subtract heat loss."""
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]

            expr = 0
            try:  # oemof.solph<=0.5.0
                expr += - m.flow[n, o, t]
                expr += m.flow[i, n, t]
            except KeyError:  # oemof.solph>=0.5.1
                period = 0  # Periods are not (yet) supported in DHNx
                expr += - m.flow[n, o, period, t]
                expr += m.flow[i, n, period, t]

            expr += - block.heat_loss[n, t]

            return expr == 0

        self.relation_no_demand = Constraint(
            self.INVESTHEATPIPES_NO_DEMAND, m.TIMESTEPS,
            rule=_relation_rule_no_demand
        )

        def _relation_rule_with_demand(block, n, t):
            """Link input and output flow and subtract heat loss."""
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]
            d = list(n.outputs.keys())[1]

            expr = 0
            try:  # oemof.solph<=0.5.0
                expr += - m.flow[n, o, t]
                expr += m.flow[i, n, t]
                expr += - block.heat_loss[n, t]
                expr += - m.flow[n, d, t]
            except KeyError:  # oemof.solph>=0.5.1
                period = 0  # Periods are not (yet) supported in DHNx
                expr += - m.flow[n, o, period, t]
                expr += m.flow[i, n, period, t]
                expr += - block.heat_loss[n, t]
                expr += - m.flow[n, d, period, t]

            return expr == 0
        self.relation_with_demand = Constraint(
            self.INVESTHEATPIPES_WITH_DEMAND, m.TIMESTEPS,
            rule=_relation_rule_with_demand
        )

        def _inflow_outflow_invest_coupling_rule(block, n):  # pylint: disable=unused-argument
            """Rule definition of constraint connecting the inflow
            `InvestmentFlowBlock.invest of pipe with invested outflow `invest`
            by nominal_storage_capacity__inflow_ratio
            """
            i = list(n.inputs.keys())[0]
            o = list(n.outputs.keys())[0]

            try:  # oemof.solph<=0.5.0
                expr = (m.InvestmentFlowBlock.invest[i, n]
                        == m.InvestmentFlowBlock.invest[n, o])
            except KeyError:  # oemof.solph>=0.5.1
                period = 0  # Periods are not (yet) supported in DHNx
                expr = (m.InvestmentFlowBlock.invest[i, n, period]
                        == m.InvestmentFlowBlock.invest[n, o, period])
            return expr
        self.inflow_outflow_invest_coupling = Constraint(
            self.INVESTHEATPIPES, rule=_inflow_outflow_invest_coupling_rule
        )



"""
Discrete DN number examples.

This examples shows how to perform an optimisation with an investment
in discrete DN numbers.

Please see the `pipes.csv` in the invest_data/network.

For each of the DN numbers a separate row is given in the table.
"""

import os

import matplotlib.pyplot as plt
import dhnx


os.chdir(os.path.dirname(__file__))

# Initialize thermal network
network = dhnx.network.ThermalNetwork()
network = network.from_csv_folder("twn_data")

# Load investment parameter
invest_opt = dhnx.input_output.load_invest_options("invest_data")

# plot network
static_map = dhnx.plotting.StaticMap(network)
static_map.draw(background_map=False)
plt.title("Given network")
plt.scatter(
    network.components.consumers["lon"],
    network.components.consumers["lat"],
    color="tab:green",
    label="consumers",
    zorder=2.5,
    s=50,
)
plt.scatter(
    network.components.producers["lon"],
    network.components.producers["lat"],
    color="tab:red",
    label="producers",
    zorder=2.5,
    s=50,
)
plt.scatter(
    network.components.forks["lon"],
    network.components.forks["lat"],
    color="tab:grey",
    label="forks",
    zorder=2.5,
    s=50,
)
plt.text(-2, 32, "P0", fontsize=14)
plt.legend()

# Execute investment optimization
network.optimize_investment(invest_options=invest_opt)

# ####### Postprocessing and Plotting ###########

# get results
results_edges = network.results.optimization["components"]["pipes"]
print(
    results_edges[
        ["from_node", "to_node", "hp_type", "capacity", "direction", "costs", "losses"]
    ]
)

# print(results_edges[['invest_costs[€]']].sum())
print("Objective value: ", network.results.optimization["oemof_meta"]["objective"])

# assign new ThermalNetwork with invested pipes
twn_results = network
twn_results.components["pipes"] = results_edges[results_edges["capacity"] > 0.001]

# plot invested edges
static_map_2 = dhnx.plotting.StaticMap(twn_results)
static_map_2.draw(background_map=False)
plt.title("Result network")
plt.scatter(
    network.components.consumers["lon"],
    network.components.consumers["lat"],
    color="tab:green",
    label="consumers",
    zorder=2.5,
    s=50,
)
plt.scatter(
    network.components.producers["lon"],
    network.components.producers["lat"],
    color="tab:red",
    label="producers",
    zorder=2.5,
    s=50,
)
plt.scatter(
    network.components.forks["lon"],
    network.components.forks["lat"],
    color="tab:grey",
    label="forks",
    zorder=2.5,
    s=50,
)
plt.text(-2, 32, "P0", fontsize=14)
for r, c in twn_results.components["pipes"].iterrows():
    size = c["hp_type"]
    type = c["from_node"].split("-")[0]
    id = c["from_node"].split("-")[1]
    lat_0 = twn_results.components[type].loc[id].lat
    lon_0 = twn_results.components[type].loc[id].lon
    type = c["to_node"].split("-")[0]
    id = c["to_node"].split("-")[1]
    lat_1 = twn_results.components[type].loc[id].lat
    lon_1 = twn_results.components[type].loc[id].lon
    lat_mid = lat_0 + 0.5 * (lat_1 - lat_0)
    lon_mid = lon_0 + 0.5 * (lon_1 - lon_0)
    plt.text(lon_mid, lat_mid, size, va="center", ha="center")

plt.legend()
plt.show()
